# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron.tests import base

from vmware_nsx.common import config
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.plugins.nsx_v import availability_zones as nsx_az


DEF_AZ_POOL = ['service:compact:1:2', 'vdr:compact:1:2']
DEF_GLOBAL_POOL = ['service:compact:4:10', 'vdr:compact:4:10']


class NsxvAvailabilityZonesTestCase(base.BaseTestCase):

    def setUp(self):
        super(NsxvAvailabilityZonesTestCase, self).setUp()
        self.az_name = 'zone1'
        self.group_name = 'az:%s' % self.az_name
        config.register_nsxv_azs(cfg.CONF, [self.az_name])
        cfg.CONF.set_override("ha_placement_random", True, group="nsxv")
        cfg.CONF.set_override("mgt_net_proxy_ips", ["2.2.2.2"], group="nsxv")
        cfg.CONF.set_override("dvs_id", "dvs-1", group="nsxv")

    def _config_az(self,
                   resource_pool_id="respool",
                   datastore_id="datastore",
                   edge_ha=True,
                   ha_datastore_id="hastore",
                   backup_edge_pool=DEF_AZ_POOL,
                   ha_placement_random=False,
                   datacenter_moid="datacenter",
                   mgt_net_moid="portgroup-407",
                   mgt_net_proxy_ips=["1.1.1.1"],
                   mgt_net_proxy_netmask="255.255.255.0",
                   mgt_net_default_gateway="2.2.2.2",
                   external_network="network-17",
                   vdn_scope_id="vdnscope-1",
                   dvs_id="dvs-2"):
        cfg.CONF.set_override("resource_pool_id", resource_pool_id,
                              group=self.group_name)
        cfg.CONF.set_override("datastore_id", datastore_id,
                              group=self.group_name)
        if edge_ha is not None:
            cfg.CONF.set_override("edge_ha", edge_ha,
                                  group=self.group_name)
        cfg.CONF.set_override("ha_datastore_id", ha_datastore_id,
                              group=self.group_name)
        if ha_placement_random is not None:
            cfg.CONF.set_override("ha_placement_random",
                                  ha_placement_random,
                                  group=self.group_name)
        if datacenter_moid is not None:
            cfg.CONF.set_override("datacenter_moid",
                                  datacenter_moid,
                                  group=self.group_name)
        if backup_edge_pool is not None:
            cfg.CONF.set_override("backup_edge_pool", backup_edge_pool,
                                  group=self.group_name)
        if mgt_net_moid is not None:
            cfg.CONF.set_override("mgt_net_moid", mgt_net_moid,
                                  group=self.group_name)
        if mgt_net_proxy_ips is not None:
            cfg.CONF.set_override("mgt_net_proxy_ips", mgt_net_proxy_ips,
                                  group=self.group_name)
        if mgt_net_proxy_netmask is not None:
            cfg.CONF.set_override("mgt_net_proxy_netmask",
                                  mgt_net_proxy_netmask,
                                  group=self.group_name)
        if mgt_net_default_gateway is not None:
            cfg.CONF.set_override("mgt_net_default_gateway",
                                  mgt_net_default_gateway,
                                  group=self.group_name)
        if external_network is not None:
            cfg.CONF.set_override("external_network", external_network,
                                  group=self.group_name)
        if vdn_scope_id is not None:
            cfg.CONF.set_override("vdn_scope_id", vdn_scope_id,
                                  group=self.group_name)
        if dvs_id is not None:
            cfg.CONF.set_override("dvs_id", dvs_id,
                                  group=self.group_name)

    def test_simple_availability_zone(self):
        self._config_az()
        az = nsx_az.NsxVAvailabilityZone(self.az_name)
        self.assertEqual(self.az_name, az.name)
        self.assertEqual("respool", az.resource_pool)
        self.assertEqual("datastore", az.datastore_id)
        self.assertTrue(az.edge_ha)
        self.assertEqual("hastore", az.ha_datastore_id)
        self.assertFalse(az.ha_placement_random)
        self.assertEqual("datacenter", az.datacenter_moid)
        self.assertEqual(DEF_AZ_POOL, az.backup_edge_pool)
        self.assertEqual("portgroup-407", az.mgt_net_moid)
        self.assertEqual(["1.1.1.1"], az.mgt_net_proxy_ips)
        self.assertEqual("255.255.255.0", az.mgt_net_proxy_netmask)
        self.assertEqual("2.2.2.2", az.mgt_net_default_gateway)
        self.assertEqual("network-17", az.external_network)
        self.assertEqual("vdnscope-1", az.vdn_scope_id)
        self.assertEqual("dvs-2", az.dvs_id)
        self.assertTrue(az.az_metadata_support)

    def test_availability_zone_no_edge_ha(self):
        self._config_az(edge_ha=False)
        az = nsx_az.NsxVAvailabilityZone(self.az_name)
        self.assertEqual(self.az_name, az.name)
        self.assertEqual("respool", az.resource_pool)
        self.assertEqual("datastore", az.datastore_id)
        self.assertFalse(az.edge_ha)
        self.assertIsNone(az.ha_datastore_id)
        self.assertFalse(az.ha_placement_random)

    def test_availability_zone_no_ha_datastore(self):
        self._config_az(ha_datastore_id=None)
        az = nsx_az.NsxVAvailabilityZone(self.az_name)
        self.assertEqual(self.az_name, az.name)
        self.assertEqual("respool", az.resource_pool)
        self.assertEqual("datastore", az.datastore_id)
        self.assertTrue(az.edge_ha)
        self.assertIsNone(az.ha_datastore_id)
        self.assertFalse(az.ha_placement_random)

    def test_missing_group_section(self):
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            "doesnt_exist")

    def test_availability_zone_missing_respool(self):
        self._config_az(resource_pool_id=None)
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            self.az_name)

    def test_availability_zone_missing_datastore(self):
        self._config_az(datastore_id=None)
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            self.az_name)

    def test_availability_zone_missing_edge_ha(self):
        self._config_az(edge_ha=None)
        az = nsx_az.NsxVAvailabilityZone(self.az_name)
        self.assertEqual(self.az_name, az.name)
        self.assertEqual("respool", az.resource_pool)
        self.assertEqual("datastore", az.datastore_id)
        self.assertFalse(az.edge_ha)
        self.assertIsNone(az.ha_datastore_id)
        self.assertFalse(az.ha_placement_random)

    def test_availability_zone_missing_edge_placement(self):
        self._config_az(ha_placement_random=None)
        az = nsx_az.NsxVAvailabilityZone(self.az_name)
        self.assertEqual(self.az_name, az.name)
        self.assertEqual("respool", az.resource_pool)
        self.assertEqual("datastore", az.datastore_id)
        self.assertTrue(az.edge_ha)
        self.assertEqual("hastore", az.ha_datastore_id)
        # ha_placement_random should have the global value
        self.assertTrue(az.ha_placement_random)

    def test_availability_zone_missing_backup_pool(self):
        self._config_az(backup_edge_pool=None)
        az = nsx_az.NsxVAvailabilityZone(self.az_name)
        self.assertEqual(self.az_name, az.name)
        # Should use the global configuration instead
        self.assertEqual(DEF_GLOBAL_POOL, az.backup_edge_pool)

    def test_availability_zone_missing_metadata(self):
        self._config_az(mgt_net_proxy_ips=None)
        az = nsx_az.NsxVAvailabilityZone(self.az_name)
        self.assertIsNone(az.mgt_net_moid)
        self.assertEqual([], az.mgt_net_proxy_ips)
        self.assertIsNone(az.mgt_net_proxy_netmask)
        self.assertIsNone(az.mgt_net_default_gateway)
        self.assertFalse(az.az_metadata_support)

    def test_availability_zone_same_metadata(self):
        self._config_az(mgt_net_proxy_ips=["2.2.2.2"])
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            self.az_name)

        self._config_az(mgt_net_proxy_ips=["2.2.2.2", "3.3.3.3"])
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            self.az_name)


class NsxvAvailabilityZonesOldTestCase(base.BaseTestCase):
    """Test old way of configuring the availability zones

    using a one-line configuration instead of different dynamic sections
    """

    def setUp(self):
        super(NsxvAvailabilityZonesOldTestCase, self).setUp()
        cfg.CONF.set_override("mgt_net_proxy_ips", ["2.2.2.2"], group="nsxv")
        cfg.CONF.set_override("dvs_id", "dvs-1", group="nsxv")

    def test_simple_availability_zone(self):
        az = nsx_az.NsxVAvailabilityZone(
            "name:respool:datastore:true:hastore")
        self.assertEqual("name", az.name)
        self.assertEqual("respool", az.resource_pool)
        self.assertEqual("datastore", az.datastore_id)
        self.assertTrue(az.edge_ha)
        self.assertEqual("hastore", az.ha_datastore_id)
        self.assertFalse(az.ha_placement_random)
        self.assertEqual(DEF_GLOBAL_POOL, az.backup_edge_pool)
        # should get the global configuration (which is empty now)
        self.assertIsNone(az.external_network)
        self.assertIsNone(az.vdn_scope_id)
        self.assertEqual("dvs-1", az.dvs_id)
        # no metadata per az support
        self.assertFalse(az.az_metadata_support)
        self.assertIsNone(az.mgt_net_moid)
        self.assertEqual([], az.mgt_net_proxy_ips)
        self.assertIsNone(az.mgt_net_proxy_netmask)
        self.assertIsNone(az.mgt_net_default_gateway)

    def test_availability_zone_without_ha_datastore(self):
        az = nsx_az.NsxVAvailabilityZone(
            "name:respool:datastore:true")
        self.assertEqual("name", az.name)
        self.assertEqual("respool", az.resource_pool)
        self.assertEqual("datastore", az.datastore_id)
        self.assertTrue(az.edge_ha)
        self.assertIsNone(az.ha_datastore_id)

    def test_availability_zone_without_edge_ha(self):
        az = nsx_az.NsxVAvailabilityZone(
            "name:respool:datastore:FALSE")
        self.assertEqual("name", az.name)
        self.assertEqual("respool", az.resource_pool)
        self.assertEqual("datastore", az.datastore_id)
        self.assertFalse(az.edge_ha)
        self.assertIsNone(az.ha_datastore_id)

    def test_availability_fail_long_name(self):
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            "very-very-very-very-very-longest-name:respool:da:true:ha")

    def test_availability_fail_few_args(self):
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            "name:respool")

    def test_availability_fail_many_args(self):
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            "name:1:2:3:4:5:6")

    def test_availability_fail_bad_edge_ha(self):
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            "name:respool:datastore:truex:hastore")

    def test_availability_fail_no_ha_datastore(self):
        self.assertRaises(
            nsx_exc.NsxInvalidConfiguration,
            nsx_az.NsxVAvailabilityZone,
            "name:respool:datastore:false:hastore")



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""nsxv3_security_groups

Revision ID: 28430956782d
Revises: 53a3254aa95e
Create Date: 2015-08-24 18:19:09.397813

"""

# revision identifiers, used by Alembic.
revision = '28430956782d'
down_revision = '53a3254aa95e'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'neutron_nsx_firewall_section_mappings',
        sa.Column('neutron_id', sa.String(36), nullable=False),
        sa.Column('nsx_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(['neutron_id'], ['securitygroups.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('neutron_id'))

    op.create_table(
        'neutron_nsx_rule_mappings',
        sa.Column('neutron_id', sa.String(36), nullable=False),
        sa.Column('nsx_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(['neutron_id'], ['securitygrouprules.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('neutron_id'))



# Copyright 2015 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import enum
import glob
import importlib
import os

from oslo_config import cfg
from oslo_log import log as logging
import requests

from vmware_nsx.common import config  # noqa
from vmware_nsx.shell.admin.plugins.common import constants

# Suppress the Insecure request warning
requests.packages.urllib3.disable_warnings()

LOG = logging.getLogger(__name__)


class Operations(enum.Enum):
    LIST = 'list'
    CLEAN = 'clean'
    CLEAN_ALL = 'clean-all'
    CREATE = 'create'
    DELETE = 'delete'
    LIST_MISMATCHES = 'list-mismatches'
    FIX_MISMATCH = 'fix-mismatch'

    NEUTRON_LIST = 'neutron-list'
    NEUTRON_CLEAN = 'neutron-clean'
    NEUTRON_UPDATE = 'neutron-update'

    NSX_LIST = 'nsx-list'
    NSX_CLEAN = 'nsx-clean'
    NSX_UPDATE = 'nsx-update'
    NSX_UPDATE_ALL = 'nsx-update-all'
    NSX_UPDATE_SECRET = 'nsx-update-secret'
    NSX_UPDATE_RULES = 'nsx-update-rules'
    NSX_UPDATE_DHCP_RELAY = 'nsx-update-dhcp-relay'
    NSX_UPDATE_IP = 'nsx-update-ip'
    NSX_RECREATE = 'nsx-recreate'
    NSX_REDISTRIBURE = 'nsx-redistribute'
    NSX_REORDER = 'nsx-reorder'
    NSX_TAG_DEFAULT = 'nsx-tag-default'
    MIGRATE_TO_DYNAMIC_CRITERIA = 'migrate-to-dynamic-criteria'
    NSX_MIGRATE_V_V3 = 'nsx-migrate-v-v3'
    MIGRATE_TO_POLICY = 'migrate-to-policy'
    NSX_MIGRATE_EXCLUDE_PORTS = 'migrate-exclude-ports'
    MIGRATE_VDR_DHCP = 'migrate-vdr-dhcp'
    STATUS = 'status'
    GENERATE = 'generate'
    IMPORT = 'import'
    SHOW = 'show'
    VALIDATE = 'validate'

ops = [op.value for op in Operations]


class Resource(object):
    def __init__(self, name, ops):
        self.name = name
        self.supported_ops = ops


# Add supported NSX-V3 resources in this dictionary
nsxv3_resources = {
    constants.SECURITY_GROUPS: Resource(constants.SECURITY_GROUPS,
                                        [Operations.LIST.value,
                                         Operations.FIX_MISMATCH.value]),
    constants.FIREWALL_SECTIONS: Resource(constants.FIREWALL_SECTIONS,
                                          [Operations.LIST.value,
                                           Operations.LIST_MISMATCHES.value]),
    constants.FIREWALL_NSX_GROUPS: Resource(
        constants.FIREWALL_NSX_GROUPS, [
            Operations.LIST.value,
            Operations.LIST_MISMATCHES.value,
            Operations.MIGRATE_TO_DYNAMIC_CRITERIA.value]),
    constants.NETWORKS: Resource(constants.NETWORKS,
                                 [Operations.LIST_MISMATCHES.value]),
    constants.PORTS: Resource(constants.PORTS,
                              [Operations.LIST_MISMATCHES.value,
                               Operations.NSX_TAG_DEFAULT.value,
                               Operations.NSX_MIGRATE_V_V3.value,
                               Operations.NSX_MIGRATE_EXCLUDE_PORTS.value]),
    constants.ROUTERS: Resource(constants.ROUTERS,
                                [Operations.LIST_MISMATCHES.value,
                                 Operations.NSX_UPDATE_RULES.value,
                                 Operations.NSX_UPDATE_DHCP_RELAY.value]),
    constants.DHCP_BINDING: Resource(constants.DHCP_BINDING,
                                     [Operations.LIST.value,
                                      Operations.NSX_UPDATE.value,
                                      Operations.NSX_RECREATE.value]),
    constants.METADATA_PROXY: Resource(constants.METADATA_PROXY,
                                       [Operations.LIST.value,
                                        Operations.NSX_UPDATE.value,
                                        Operations.NSX_UPDATE_IP.value]),
    constants.ORPHANED_DHCP_SERVERS: Resource(constants.ORPHANED_DHCP_SERVERS,
                                              [Operations.NSX_LIST.value,
                                               Operations.NSX_CLEAN.value]),
    constants.CERTIFICATE: Resource(constants.CERTIFICATE,
                                    [Operations.GENERATE.value,
                                     Operations.SHOW.value,
                                     Operations.CLEAN.value,
                                     Operations.IMPORT.value,
                                     Operations.NSX_LIST.value]),
    constants.CONFIG: Resource(constants.CONFIG,
                               [Operations.VALIDATE.value]),
    constants.ORPHANED_NETWORKS: Resource(constants.ORPHANED_NETWORKS,
                                 [Operations.LIST.value,
                                  Operations.NSX_CLEAN.value]),
    constants.ORPHANED_ROUTERS: Resource(constants.ORPHANED_ROUTERS,
                                [Operations.LIST.value,
                                 Operations.NSX_CLEAN.value]),
    constants.LB_SERVICES: Resource(constants.LB_SERVICES,
                                    [Operations.LIST.value]),
    constants.LB_VIRTUAL_SERVERS: Resource(constants.LB_VIRTUAL_SERVERS,
                                           [Operations.LIST.value]),
    constants.LB_POOLS: Resource(constants.LB_POOLS,
                                 [Operations.LIST.value]),
    constants.LB_MONITORS: Resource(constants.LB_MONITORS,
                                    [Operations.LIST.value]),
    constants.RATE_LIMIT: Resource(constants.RATE_LIMIT,
                                   [Operations.SHOW.value,
                                    Operations.NSX_UPDATE.value])
}

# Add supported NSX-V resources in this dictionary
nsxv_resources = {
    constants.EDGES: Resource(constants.EDGES,
                              [Operations.NSX_LIST.value,
                               Operations.NEUTRON_LIST.value,
                               Operations.NSX_UPDATE.value,
                               Operations.NSX_UPDATE_ALL.value]),
    constants.BACKUP_EDGES: Resource(constants.BACKUP_EDGES,
                                     [Operations.LIST.value,
                                      Operations.CLEAN.value,
                                      Operations.CLEAN_ALL.value,
                                      Operations.LIST_MISMATCHES.value,
                                      Operations.FIX_MISMATCH.value,
                                      Operations.NEUTRON_CLEAN.value]),
    constants.ORPHANED_EDGES: Resource(constants.ORPHANED_EDGES,
                                       [Operations.LIST.value,
                                        Operations.CLEAN.value]),
    constants.ORPHANED_BINDINGS: Resource(constants.ORPHANED_BINDINGS,
                                          [Operations.LIST.value,
                                           Operations.CLEAN.value]),
    constants.MISSING_EDGES: Resource(constants.MISSING_EDGES,
                                      [Operations.LIST.value]),
    constants.SPOOFGUARD_POLICY: Resource(constants.SPOOFGUARD_POLICY,
                                          [Operations.LIST.value,
                                           Operations.CLEAN.value]),
    constants.DHCP_BINDING: Resource(constants.DHCP_BINDING,
                                     [Operations.LIST.value,
                                      Operations.NSX_UPDATE.value,
                                      Operations.NSX_REDISTRIBURE.value,
                                      Operations.NSX_RECREATE.value]),
    constants.NETWORKS: Resource(constants.NETWORKS,
                                 [Operations.LIST.value,
                                  Operations.NSX_UPDATE.value]),
    constants.MISSING_NETWORKS: Resource(constants.MISSING_NETWORKS,
                                [Operations.LIST.value]),
    constants.ORPHANED_NETWORKS: Resource(constants.ORPHANED_NETWORKS,
                                 [Operations.LIST.value,
                                  Operations.NSX_CLEAN.value]),
    constants.SECURITY_GROUPS: Resource(constants.SECURITY_GROUPS,
                                        [Operations.LIST.value,
                                         Operations.FIX_MISMATCH.value,
                                         Operations.MIGRATE_TO_POLICY.value]),
    constants.FIREWALL_NSX_GROUPS: Resource(
        constants.FIREWALL_NSX_GROUPS, [Operations.LIST.value,
                                        Operations.LIST_MISMATCHES.value]),
    constants.FIREWALL_SECTIONS: Resource(constants.FIREWALL_SECTIONS,
                                          [Operations.LIST.value,
                                           Operations.LIST_MISMATCHES.value,
                                           Operations.NSX_UPDATE.value,
                                           Operations.NSX_REORDER.value]),
    constants.METADATA: Resource(
        constants.METADATA, [Operations.NSX_UPDATE.value,
                             Operations.NSX_UPDATE_SECRET.value,
                             Operations.STATUS.value]),
    constants.ROUTERS: Resource(constants.ROUTERS,
                                [Operations.NSX_RECREATE.value,
                                 Operations.NSX_REDISTRIBURE.value,
                                 Operations.MIGRATE_VDR_DHCP.value]),
    constants.ORPHANED_VNICS: Resource(constants.ORPHANED_VNICS,
                                       [Operations.NSX_LIST.value,
                                        Operations.NSX_CLEAN.value]),
    constants.CONFIG: Resource(constants.CONFIG,
                               [Operations.VALIDATE.value]),
    constants.BGP_GW_EDGE: Resource(constants.BGP_GW_EDGE,
                                    [Operations.CREATE.value,
                                     Operations.DELETE.value,
                                     Operations.LIST.value]),
    constants.ROUTING_REDIS_RULE: Resource(constants.ROUTING_REDIS_RULE,
                                           [Operations.CREATE.value,
                                            Operations.DELETE.value]),
    constants.BGP_NEIGHBOUR: Resource(constants.BGP_NEIGHBOUR,
                                      [Operations.CREATE.value,
                                       Operations.DELETE.value])
}


# Add supported NSX-TVD resources in this dictionary
# TODO(asarfaty): add v+v3 resources here too
nsxtvd_resources = {
    constants.PROJECTS: Resource(constants.PROJECTS,
                                 [Operations.IMPORT.value]),
}

nsxv3_resources_names = list(nsxv3_resources.keys())
nsxv_resources_names = list(nsxv_resources.keys())
nsxtvd_resources_names = list(nsxtvd_resources.keys())


def get_resources(plugin_dir):
    modules = glob.glob(plugin_dir + "/*.py")
    return map(lambda module: os.path.splitext(os.path.basename(module))[0],
               modules)


def get_plugin():
    plugin = cfg.CONF.core_plugin
    plugin_name = ''
    if plugin in (constants.NSXV3_PLUGIN, constants.VMWARE_NSXV3):
        plugin_name = 'nsxv3'
    elif plugin in (constants.NSXV_PLUGIN, constants.VMWARE_NSXV):
        plugin_name = 'nsxv'
    elif plugin in (constants.NSXTVD_PLUGIN, constants.VMWARE_NSXTVD):
        plugin_name = 'nsxtvd'
    return plugin_name


def _get_choices():
    plugin = get_plugin()
    if plugin == 'nsxv3':
        return nsxv3_resources_names
    elif plugin == 'nsxv':
        return nsxv_resources_names
    elif plugin == 'nsxtvd':
        return nsxtvd_resources_names


def _get_resources():
    plugin = get_plugin()
    if plugin == 'nsxv3':
        return 'NSX-V3 resources: %s' % (', '.join(nsxv3_resources_names))
    elif plugin == 'nsxv':
        return 'NSX-V resources: %s' % (', '.join(nsxv_resources_names))
    elif plugin == 'nsxtvd':
        return 'NSX-TVD resources: %s' % (', '.join(nsxtvd_resources_names))


cli_opts = [cfg.StrOpt('fmt',
                       short='f',
                       default='psql',
                       choices=['psql', 'json'],
                       help='Supported output formats: json, psql'),
            cfg.StrOpt('resource',
                       short='r',
                       choices=_get_choices(),
                       help=_get_resources()),
            cfg.StrOpt('operation',
                       short='o',
                       help='Supported list of operations: {}'
                             .format(', '.join(ops))),
            cfg.StrOpt('plugin',
                       help='nsxv or nsxv3 if the tvd plugin is used'),
            cfg.BoolOpt('force',
                        default=False,
                        help='Enables \'force\' mode. No confirmations will '
                             'be made before deletions.'),
            cfg.MultiStrOpt('property',
                            short='p',
                            help='Key-value pair containing the information '
                                 'to be updated. For ex: key=value.'),
            cfg.BoolOpt('verbose',
                        short='v',
                        default=False,
                        help='Triggers detailed output for some commands')
            ]


# Describe dependencies between admin utils resources and external libraries
# that are not always installed
resources_dependencies = {
    'nsxv': {'gw_edges': ['neutron_dynamic_routing.extensions']}}


def verify_external_dependencies(plugin_name, resource):
    if plugin_name in resources_dependencies:
        deps = resources_dependencies[plugin_name]
        if resource in deps:
            for d in deps[resource]:
                try:
                    importlib.import_module(d)
                except ImportError:
                    return False
    return True


def init_resource_plugin(plugin_name, plugin_dir):
    plugin_resources = get_resources(plugin_dir)
    for resource in plugin_resources:
        if (resource != '__init__'):
            # skip unsupported resources
            if not verify_external_dependencies(plugin_name, resource):
                LOG.info("Skipping resource %s because of dependencies",
                         resource)
                continue
            # load the resource
            importlib.import_module(
                "vmware_nsx.shell.admin.plugins."
                "{}.resources.".format(plugin_name) + resource)


def get_plugin_dir(plugin_name):
    plugin_dir = (os.path.dirname(os.path.realpath(__file__)) +
                  "/admin/plugins")
    return '{}/{}/resources'.format(plugin_dir, plugin_name)



# Copyright 2016 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.objects.qos import policy as qos_policy
from neutron_lib.services.qos import constants as qos_consts


def update_network_policy_binding(context, net_id, new_policy_id):
    # detach the old policy (if exists) from the network
    old_policy = qos_policy.QosPolicy.get_network_policy(
        context, net_id)
    if old_policy:
        if old_policy.id == new_policy_id:
            return
        old_policy.detach_network(net_id)

    # attach the new policy (if exists) to the network
    if new_policy_id is not None:
        new_policy = qos_policy.QosPolicy.get_object(
            context, id=new_policy_id)
        if new_policy:
            new_policy.attach_network(net_id)


def update_port_policy_binding(context, port_id, new_policy_id):
    # detach the old policy (if exists) from the port
    old_policy = qos_policy.QosPolicy.get_port_policy(
        context, port_id)
    if old_policy:
        if old_policy.id == new_policy_id:
            return
        old_policy.detach_port(port_id)

    # attach the new policy (if exists) to the port
    if new_policy_id is not None:
        new_policy = qos_policy.QosPolicy.get_object(
            context, id=new_policy_id)
        if new_policy:
            new_policy.attach_port(port_id)


def get_port_policy_id(context, port_id):
    policy = qos_policy.QosPolicy.get_port_policy(
        context, port_id)
    if policy:
        return policy.id


def get_network_policy_id(context, net_id):
    policy = qos_policy.QosPolicy.get_network_policy(
        context, net_id)
    if policy:
        return policy.id


def set_qos_policy_on_new_net(context, net_data, created_net):
    """Update the network with the assigned or default QoS policy

    Update the network-qos binding table, and the new network structure
    """
    qos_policy_id = net_data.get(qos_consts.QOS_POLICY_ID)
    if not qos_policy_id:
        # try and get the default one
        qos_obj = qos_policy.QosPolicyDefault.get_object(
            context, project_id=created_net['project_id'])
        if qos_obj:
            qos_policy_id = qos_obj.qos_policy_id

    if qos_policy_id:
        # attach the policy to the network in the neutron DB
        update_network_policy_binding(
            context,
            net_data['id'],
            qos_policy_id)
    created_net[qos_consts.QOS_POLICY_ID] = qos_policy_id
    return qos_policy_id



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import locking
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_v3 import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgeMemberManager(base_mgr.Nsxv3LoadbalancerBaseManager):
    @log_helpers.log_method_call
    def __init__(self):
        super(EdgeMemberManager, self).__init__()

    @log_helpers.log_method_call
    def _get_info_from_fip(self, context, fip):
        filters = {'floating_ip_address': [fip]}
        floating_ips = self.core_plugin.get_floatingips(context,
                                                        filters=filters)
        if floating_ips:
            return (floating_ips[0]['fixed_ip_address'],
                    floating_ips[0]['router_id'])
        else:
            msg = (_('Cannot get floating ip %(fip)s provided from '
                     'neutron db') % {'fip': fip})
            raise n_exc.BadRequest(resource='lbaas-vip', msg=msg)

    @log_helpers.log_method_call
    def _create_lb_service(self, context, service_client, tenant_id,
                           router_id, nsx_router_id, lb_id, lb_size):
        router = self.core_plugin.get_router(context, router_id)
        if not router.get('external_gateway_info'):
            msg = (_('Tenant router %(router)s does not connect to '
                     'external gateway') % {'router': router['id']})
            raise n_exc.BadRequest(resource='lbaas-lbservice-create',
                                   msg=msg)
        lb_name = utils.get_name_and_uuid(router['name'] or 'router',
                                          router_id)
        tags = lb_utils.get_tags(self.core_plugin, router_id,
                                 lb_const.LR_ROUTER_TYPE,
                                 tenant_id, context.project_name)
        attachment = {'target_id': nsx_router_id,
                      'target_type': 'LogicalRouter'}
        lb_service = service_client.create(display_name=lb_name,
                                           tags=tags,
                                           attachment=attachment,
                                           size=lb_size)
        # Update router to enable advertise_lb_vip flag
        self.core_plugin.nsxlib.logical_router.update_advertisement(
            nsx_router_id, advertise_lb_vip=True)
        return lb_service

    def _get_updated_pool_members(self, context, lb_pool, member):
        network = lb_utils.get_network_from_subnet(
            context, self.core_plugin, member.subnet_id)
        if network.get('router:external'):
            fixed_ip, router_id = self._get_info_from_fip(
                context, member.address)
        else:
            fixed_ip = member.address
        for m in lb_pool['members']:
            if m['ip_address'] == fixed_ip:
                m['display_name'] = member.name[:219] + '_' + member.id
                m['weight'] = member.weight
        return lb_pool['members']

    @log_helpers.log_method_call
    def _add_loadbalancer_binding(self, context, lb_id, lbs_id,
                                  nsx_router_id, vip_address):
        # First check if there is already binding for the lb.
        # If there is no binding for the lb, add the db binding.
        binding = nsx_db.get_nsx_lbaas_loadbalancer_binding(
            context.session, lb_id)
        if not binding:
            nsx_db.add_nsx_lbaas_loadbalancer_binding(
                context.session, lb_id, lbs_id,
                nsx_router_id, vip_address)
        else:
            LOG.debug("LB binding has already been added, and no need "
                      "to add here.")

    @log_helpers.log_method_call
    def create(self, context, member):
        with locking.LockManager.get_lock('member-%s' %
                                          str(member.pool.loadbalancer_id)):
            self._member_create(context, member)

    def _member_create(self, context, member):
        lb_id = member.pool.loadbalancer_id
        pool_id = member.pool.id
        loadbalancer = member.pool.loadbalancer
        if not lb_utils.validate_lb_subnet(context, self.core_plugin,
                                           member.subnet_id):
            msg = (_('Cannot add member %(member)s to pool as member subnet '
                     '%(subnet)s is neither public nor connected to router') %
                   {'member': member.id, 'subnet': member.subnet_id})
            raise n_exc.BadRequest(resource='lbaas-subnet', msg=msg)

        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        service_client = self.core_plugin.nsxlib.load_balancer.service

        network = lb_utils.get_network_from_subnet(
            context, self.core_plugin, member.subnet_id)
        if network.get('router:external'):
            router_id, fixed_ip = self._get_info_from_fip(
                context, member.address)
        else:
            router_id = lb_utils.get_router_from_network(
                context, self.core_plugin, member.subnet_id)
            fixed_ip = member.address

        binding = nsx_db.get_nsx_lbaas_pool_binding(context.session,
                                                    lb_id, pool_id)
        if binding:
            vs_id = binding.get('lb_vs_id')
            lb_pool_id = binding.get('lb_pool_id')
            lb_binding = nsx_db.get_nsx_lbaas_loadbalancer_binding(
                context.session, lb_id)
            if not lb_binding:
                nsx_router_id = nsx_db.get_nsx_router_id(context.session,
                                                         router_id)
                lb_service = service_client.get_router_lb_service(
                    nsx_router_id)
                if not lb_service:
                    lb_size = lb_utils.get_lb_flavor_size(
                        self.flavor_plugin, context, loadbalancer.flavor_id)
                    lb_service = self._create_lb_service(
                        context, service_client, member.tenant_id,
                        router_id, nsx_router_id, loadbalancer.id, lb_size)
                if lb_service:
                    lb_service_id = lb_service['id']
                    self._add_loadbalancer_binding(
                        context, loadbalancer.id, lb_service_id,
                        nsx_router_id, loadbalancer.vip_address)
                    if vs_id:
                        try:
                            service_client.add_virtual_server(lb_service_id,
                                                              vs_id)
                        except nsxlib_exc.ManagerError:
                            self.lbv2_driver.member.failed_completion(context,
                                                                      member)
                            msg = (_('Failed to attach virtual server %(vs)s '
                                   'to lb service %(service)s') %
                                   {'vs': vs_id, 'service': lb_service_id})
                            raise n_exc.BadRequest(resource='lbaas-member',
                                                   msg=msg)
                else:
                    msg = (_('Failed to get lb service to attach virtual '
                             'server %(vs)s for member %(member)s') %
                           {'vs': vs_id, 'member': member['id']})
                    raise nsx_exc.NsxPluginException(err_msg=msg)

            lb_pool = pool_client.get(lb_pool_id)
            old_m = lb_pool.get('members', None)
            new_m = [{'display_name': member.name[:219] + '_' + member.id,
                      'ip_address': fixed_ip,
                      'port': member.protocol_port,
                      'weight': member.weight}]
            members = (old_m + new_m) if old_m else new_m
            pool_client.update_pool_with_members(lb_pool_id, members)
        else:
            msg = (_('Failed to get pool binding to add member %s') %
                   member['id'])
            raise nsx_exc.NsxPluginException(err_msg=msg)

        self.lbv2_driver.member.successful_completion(context, member)

    @log_helpers.log_method_call
    def update(self, context, old_member, new_member):
        lb_id = old_member.pool.loadbalancer_id
        pool_id = old_member.pool.id
        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, pool_id)
        if pool_binding:
            lb_pool_id = pool_binding.get('lb_pool_id')
            try:
                lb_pool = pool_client.get(lb_pool_id)
                updated_members = self._get_updated_pool_members(
                    context, lb_pool, new_member)
                pool_client.update_pool_with_members(lb_pool_id,
                                                     updated_members)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    self.lbv2_driver.member.failed_completion(
                        context, new_member)
                    LOG.error('Failed to update member %(member)s: '
                              '%(err)s',
                              {'member': old_member.id, 'err': e})
        self.lbv2_driver.member.successful_completion(
            context, new_member)

    @log_helpers.log_method_call
    def delete(self, context, member):
        lb_id = member.pool.loadbalancer_id
        pool_id = member.pool.id
        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, pool_id)
        if pool_binding:
            lb_pool_id = pool_binding.get('lb_pool_id')
            try:
                lb_pool = pool_client.get(lb_pool_id)
                network = lb_utils.get_network_from_subnet(
                    context, self.core_plugin, member.subnet_id)
                if network.get('router:external'):
                    fixed_ip, router_id = self._get_info_from_fip(
                        context, member.address)
                else:
                    fixed_ip = member.address
                if 'members' in lb_pool:
                    m_list = lb_pool['members']
                    members = [m for m in m_list
                               if m['ip_address'] != fixed_ip]
                    pool_client.update_pool_with_members(lb_pool_id, members)
            except nsxlib_exc.ManagerError:
                self.lbv2_driver.member.failed_completion(context, member)
                msg = _('Failed to remove member from pool on NSX backend')
                raise n_exc.BadRequest(resource='lbaas-member', msg=msg)
        self.lbv2_driver.member.successful_completion(
            context, member, delete=True)



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import mock
from openstackclient.tests.unit.network.v2 import test_subnet
from openstackclient.tests.unit import utils as tests_utils

from vmware_nsx.osc.v2 import subnet


supported_extensions = ('dhcp-mtu', 'dns-search-domain')


class TestCreateSubnet(test_subnet.TestCreateSubnet):

    def setUp(self):
        super(TestCreateSubnet, self).setUp()
        # Get the command object to test
        self.cmd = subnet.NsxCreateSubnet(self.app, self.namespace)
        # mock the relevant extensions
        get_ext = mock.patch('vmware_nsx.osc.v2.utils.get_extensions').start()
        get_ext.return_value = supported_extensions

    def _test_create_with_arg_and_val(self, arg_name, arg_val, is_valid=True):
        self.network.create_subnet = mock.Mock(return_value=self._subnet)
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            "--subnet-range", self._subnet.cidr,
            "--network", self._subnet.network_id,
            conv_name, str(arg_val),
            self._subnet.name
        ]
        verifylist = [
            ('name', self._subnet.name),
            ('subnet_range', self._subnet.cidr),
            ('network', self._subnet.network_id),
            ('ip_version', self._subnet.ip_version),
            ('gateway', 'auto'),
            (arg_name, arg_val),
        ]
        if not is_valid:
            self.assertRaises(tests_utils.ParserException, self.check_parser,
                              self.cmd, arglist, verifylist)
            return

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.network.create_subnet.assert_called_once_with(**{
            'cidr': mock.ANY,
            'ip_version': mock.ANY,
            'network_id': mock.ANY,
            'name': self._subnet.name,
            arg_name: arg_val,
        })

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def _test_create_with_tag(self, add_tags=True):
        self.skipTest('Unblock gate')

    def _test_create_with_mtu(self, mtu, is_valid=True):
        self._test_create_with_arg_and_val('dhcp_mtu', mtu, is_valid)

    def test_create_with_mtu(self):
        # check a valid value
        self._test_create_with_mtu(1500)

    def test_create_with_illegal_mtu(self):
        self._test_create_with_mtu('illegal', is_valid=False)

    def _test_create_with_search_domain(self, val, is_valid=True):
        self._test_create_with_arg_and_val('dns_search_domain', val, is_valid)

    def test_create_with_search_domain(self):
        # check a valid value
        self._test_create_with_search_domain('www.aaa.com')

    # Cannot check illegal search domain - validation is on the server side


class TestSetSubnet(test_subnet.TestSetSubnet):

    def setUp(self):
        super(TestSetSubnet, self).setUp()
        # Get the command object to test
        self.cmd = subnet.NsxSetSubnet(self.app, self.namespace)
        # mock the relevant extensions
        get_ext = mock.patch('vmware_nsx.osc.v2.utils.get_extensions').start()
        get_ext.return_value = supported_extensions

    def _test_set_with_arg_and_val(self, arg_name, arg_val, is_valid=True):
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            conv_name, str(arg_val),
            self._subnet.name,
        ]
        verifylist = [
            (arg_name, arg_val),
            ('subnet', self._subnet.name),
        ]

        if not is_valid:
            self.assertRaises(tests_utils.ParserException, self.check_parser,
                              self.cmd, arglist, verifylist)
            return

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            arg_name: arg_val
        }
        self.network.update_subnet.assert_called_with(self._subnet, **attrs)
        self.assertIsNone(result)

    def _test_set_mtu(self, mtu, is_valid=True):
        self._test_set_with_arg_and_val('dhcp_mtu', mtu, is_valid)

    def test_set_mtu(self):
        # check a valid value
        self._test_set_mtu(1500)

    def test_set_illegal_mtu(self):
        self._test_set_mtu('illegal', is_valid=False)

    def _test_set_with_search_domain(self, val, is_valid=True):
        self._test_set_with_arg_and_val('dns_search_domain', val, is_valid)

    def test_set_with_search_domain(self):
        # check a valid value
        self._test_set_with_search_domain('www.aaa.com')



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as n_context
from neutron_lib.exceptions import firewall_v1 as exceptions
from neutron_lib.plugins import directory
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron_fwaas.services.firewall.drivers import fwaas_base

from vmware_nsx.common import locking
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.nsx_v.vshield import edge_utils

LOG = logging.getLogger(__name__)
FWAAS_DRIVER_NAME = 'Fwaas V1 NSX-V driver'
RULE_NAME_PREFIX = 'Fwaas-'


class EdgeFwaasDriver(fwaas_base.FwaasDriverBase):
    """NSX-V driver for Firewall As A Service - V1."""

    @property
    def core_plugin(self):
        if not self._core_plugin:
            self._core_plugin = directory.get_plugin()
            if self._core_plugin.is_tvd_plugin():
                self._core_plugin = self._core_plugin.get_plugin_by_type(
                    projectpluginmap.NsxPlugins.NSX_V)
            # make sure plugin init was completed
            if not self._core_plugin.init_is_complete:
                self._core_plugin.init_complete(None, None, {})
        return self._core_plugin

    @property
    def edge_manager(self):
        return self.core_plugin.edge_manager

    def __init__(self):
        LOG.debug("Loading FWaaS V1 NsxVDriver.")
        super(EdgeFwaasDriver, self).__init__()
        self.driver_name = FWAAS_DRIVER_NAME
        self._core_plugin = None

    def should_apply_firewall_to_router(self, router_data,
                                        raise_exception=True):
        """Return True if the firewall rules should be added the router

        Return False in those cases:
        - router without an external gateway (rule may be added later when
                                              there is a gateway)

        Raise an exception if the router is unsupported
        (and raise_exception is True):
        - shared router (not supported)
        - md proxy router (not supported)

        """
        if (not router_data.get('distributed') and
            router_data.get('router_type') == 'shared'):
            LOG.error("Cannot apply firewall to shared router %s",
                      router_data['id'])
            if raise_exception:
                raise exceptions.FirewallInternalDriverError(
                    driver=self.driver_name)
            return False

        if router_data.get('name', '').startswith('metadata_proxy_router'):
            LOG.error("Cannot apply firewall to the metadata proxy router %s",
                      router_data['id'])
            if raise_exception:
                raise exceptions.FirewallInternalDriverError(
                    driver=self.driver_name)
            return False

        if not router_data.get('external_gateway_info'):
            LOG.info("Cannot apply firewall to router %s with no gateway",
                     router_data['id'])
            return False

        return True

    def _get_routers_edges(self, context, apply_list):
        # Get edges for all the routers in the apply list.
        # note that shared routers are currently not supported
        edge_manager = self.edge_manager
        edges_map = {}
        for router_info in apply_list:

            # No FWaaS rules needed if there is no external gateway
            if not self.should_apply_firewall_to_router(router_info.router):
                continue

            lookup_id = None
            router_id = router_info.router_id
            if router_info.router.get('distributed'):
                # Distributed router
                # we need the plr edge id
                lookup_id = edge_manager.get_plr_by_tlr_id(
                    context, router_id)
            else:
                # Exclusive router
                lookup_id = router_id
            if lookup_id:
                # look for the edge id in the DB
                edge_id = edge_utils.get_router_edge_id(context, lookup_id)
                if edge_id:
                    edges_map[router_id] = {'edge_id': edge_id,
                                            'lookup_id': lookup_id}
        return edges_map

    def _translate_rules(self, fwaas_rules, logged=False):
        translated_rules = []
        for rule in fwaas_rules:
            if not rule['enabled']:
                # skip disabled rules
                continue
            # Make sure the rule has a name, and it starts with the prefix
            # (backend max name length is 30)
            if rule.get('name'):
                rule['name'] = RULE_NAME_PREFIX + rule['name']
            else:
                rule['name'] = RULE_NAME_PREFIX + rule['id']
            rule['name'] = rule['name'][:30]
            # source & destination should be lists
            if rule.get('destination_ip_address'):
                rule['destination_ip_address'] = [
                    rule['destination_ip_address']]
            if rule.get('source_ip_address'):
                rule['source_ip_address'] = [rule['source_ip_address']]
            if logged:
                rule['logged'] = True
            translated_rules.append(rule)

        return translated_rules

    def _set_rules_on_router_edge(self, context, router_id, neutron_id,
                                  edge_id, fw_id, translated_rules,
                                  delete_fw=False):
        """Recreate router edge firewall rules

        Using the plugin code to recreate all the rules with the additional
        FWaaS rules.

        router_id is the is of the router about to be updated
            (in case of distributed router - the plr)
        neutron_id is the neutron router id
        """
        # update the backend
        router_db = self.core_plugin._get_router(context, neutron_id)
        try:
            with locking.LockManager.get_lock(str(edge_id)):
                self.core_plugin.update_router_firewall(
                    context, router_id, router_db,
                    fwaas_rules=translated_rules)
        except Exception as e:
            # catch known library exceptions and raise Fwaas generic exception
            LOG.error("Failed to update firewall %(fw)s on edge %(edge_id)s: "
                      "%(e)s", {'e': e, 'fw': fw_id, 'edge_id': edge_id})
            raise exceptions.FirewallInternalDriverError(
                driver=self.driver_name)

    def _create_or_update_firewall(self, agent_mode, apply_list, firewall):
        # admin state down means default block rule firewall
        if not firewall['admin_state_up']:
            self.apply_default_policy(agent_mode, apply_list, firewall)
            return

        # get router-edge mapping
        context = n_context.get_admin_context()
        edges_map = self._get_routers_edges(context, apply_list)
        if not edges_map:
            routers = [r.router_id for r in apply_list]
            LOG.warning("Cannot apply the firewall %(fw)s to any of the "
                        "routers %(rtrs)s",
                        {'fw': firewall['id'], 'rtrs': routers})
            return

        # Translate the FWaaS rules
        # TODO(asarfaty): get this value from the firewall extensions
        logged = False
        rules = self._translate_rules(firewall['firewall_rule_list'],
                                      logged=logged)

        # update each relevant edge with the new rules
        for router_info in apply_list:
            neutron_id = router_info.router_id
            info = edges_map.get(neutron_id)
            if info:
                self._set_rules_on_router_edge(
                    context, info['lookup_id'], neutron_id, info['edge_id'],
                    firewall['id'], rules)

    @log_helpers.log_method_call
    def create_firewall(self, agent_mode, apply_list, firewall):
        """Create the Firewall with a given policy. """
        self._create_or_update_firewall(agent_mode, apply_list, firewall)

    @log_helpers.log_method_call
    def update_firewall(self, agent_mode, apply_list, firewall):
        """Remove previous policy and apply the new policy."""
        self._create_or_update_firewall(agent_mode, apply_list, firewall)

    def _delete_firewall_or_set_default_policy(self, apply_list, firewall,
                                               delete_fw=False):
        # get router-edge mapping
        context = n_context.get_admin_context()
        edges_map = self._get_routers_edges(context, apply_list)

        # if the firewall is deleted, rules should be None
        rules = None if delete_fw else []

        # Go over all routers and update them on backend
        for router_info in apply_list:
            neutron_id = router_info.router_id
            info = edges_map.get(neutron_id)
            if info:
                self._set_rules_on_router_edge(
                    context, info['lookup_id'], neutron_id, info['edge_id'],
                    firewall['id'], rules, delete_fw=delete_fw)

    @log_helpers.log_method_call
    def delete_firewall(self, agent_mode, apply_list, firewall):
        """Delete firewall.

        Removes rules created by this instance from the backend firewall
        And add the default allow-external rule.
        """
        self._delete_firewall_or_set_default_policy(apply_list, firewall,
                                                    delete_fw=True)

    @log_helpers.log_method_call
    def apply_default_policy(self, agent_mode, apply_list, firewall):
        """Apply the default policy (deny all).

        The backend firewall always has this policy (=deny all) as default,
        so we only need to delete the current rules.
        """
        self._delete_firewall_or_set_default_policy(apply_list, firewall,
                                                    delete_fw=False)

    def get_firewall_translated_rules(self, firewall):
        if firewall['admin_state_up']:
            # TODO(asarfaty): get this value from the firewall extensions
            logged = False
            return self._translate_rules(firewall['firewall_rule_list'],
                                         logged=logged)
        return []



# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron_lib import context as n_context
from neutron_lib import exceptions
from neutron_lib.plugins import directory

from vmware_nsx.db import db as nsx_db


def is_tvd_core_plugin():
    core_plugin = cfg.CONF.core_plugin
    if (core_plugin.endswith('NsxTVDPlugin') or
        core_plugin.endswith('vmware_nsxtvd')):
        return True
    return False


def get_tvd_plugin_type_for_project(project_id, context=None):
    """Get the plugin type used by a project

    Raise an exception if not found or the plugin is not in use
    """
    if not context:
        context = n_context.get_admin_context()
    core_plugin = directory.get_plugin()
    return core_plugin.get_plugin_type_from_project(context, project_id)


def filter_plugins(cls):
    """
    Class decorator to separate the results of each of the given methods
    by plugin
    """
    def get_project_mapping(context, project_id):
        """Return the plugin associated with this project"""
        mapping = nsx_db.get_project_plugin_mapping(
                context.session, project_id)
        if mapping:
            return mapping['plugin']
        else:
            raise exceptions.ObjectNotFound(id=project_id)

    def add_separate_plugin_hook(name):
        orig_method = getattr(cls, name, None)

        def filter_results_by_plugin(self, context, **kwargs):
            """Run the original get-list method, and filter the results
            by the project id of the context
            """
            entries = orig_method(self, context, **kwargs)
            if not context.project_id:
                return entries
            req_p = get_project_mapping(context, context.project_id)
            for entry in entries[:]:
                if entry.get('tenant_id'):
                    p = get_project_mapping(context, entry['tenant_id'])
                    if p != req_p:
                        entries.remove(entry)

            return entries

        setattr(cls, name, filter_results_by_plugin)

    for method in cls.methods_to_separate:
        add_separate_plugin_hook(method)

    return cls



# Copyright 2014 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_vmware import vim_util

from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.dvs import dvs_utils

LOG = logging.getLogger(__name__)
PORTGROUP_PREFIX = 'dvportgroup'
API_FIND_ALL_BY_UUID = 'FindAllByUuid'

# QoS related constants
QOS_IN_DIRECTION = 'incomingPackets'
QOS_AGENT_NAME = 'dvfilter-generic-vmware'
DSCP_RULE_DESCRIPTION = 'Openstack Dscp Marking RUle'


class SingleDvsManager(object):
    """Management class for dvs related tasks for the dvs plugin

    For the globally configured dvs.
    the moref of the configured DVS will be learnt. This will be used in
    the operations supported by the manager.
    """
    def __init__(self):
        self._dvs = DvsManager()
        self._dvs_moref = self._get_dvs_moref_by_name(
            self._dvs.get_vc_session(),
            dvs_utils.dvs_name_get())

    def _get_dvs_moref_by_name(self, session, dvs_name):
        """Get the moref of the configured DVS."""
        results = session.invoke_api(vim_util,
                                     'get_objects',
                                     session.vim,
                                     'DistributedVirtualSwitch',
                                     100)
        while results:
            for dvs in results.objects:
                for prop in dvs.propSet:
                    if dvs_name == prop.val:
                        vim_util.cancel_retrieval(session.vim, results)
                        return dvs.obj
            results = vim_util.continue_retrieval(session.vim, results)
        raise nsx_exc.DvsNotFound(dvs=dvs_name)

    def add_port_group(self, net_id, vlan_tag=None, trunk_mode=False):
        return self._dvs.add_port_group(self._dvs_moref, net_id,
                                        vlan_tag=vlan_tag,
                                        trunk_mode=trunk_mode)

    def delete_port_group(self, net_id):
        return self._dvs.delete_port_group(self._dvs_moref, net_id)

    def get_port_group_info(self, net_id):
        return self._dvs.get_port_group_info(self._dvs_moref, net_id)

    def net_id_to_moref(self, net_id):
        return self._dvs._net_id_to_moref(self._dvs_moref, net_id)


class VCManagerBase(object):
    """Base class for all VC related classes, to initialize the session"""
    def __init__(self):
        """Initializer.

        A global session with the VC will be established.

        NOTE: the DVS port group name will be the Neutron network UUID.
        """
        self._session = dvs_utils.dvs_create_session()

    def get_vc_session(self):
        return self._session


class DvsManager(VCManagerBase):
    """Management class for dvs related tasks

    The dvs-id is not a class member, ince multiple dvs-es can be supported.
    """

    def _get_dvs_moref_by_id(self, dvs_id):
        return vim_util.get_moref(dvs_id, 'VmwareDistributedVirtualSwitch')

    def _get_vlan_spec(self, vlan_tag):
        """Gets portgroup vlan spec."""
        # Create the spec for the vlan tag
        client_factory = self._session.vim.client.factory
        spec_ns = 'ns0:VmwareDistributedVirtualSwitchVlanIdSpec'
        vl_spec = client_factory.create(spec_ns)
        vl_spec.vlanId = vlan_tag
        vl_spec.inherited = '0'
        return vl_spec

    def _get_trunk_vlan_spec(self, start=0, end=4094):
        """Gets portgroup trunk vlan spec."""
        client_factory = self._session.vim.client.factory
        spec_ns = 'ns0:VmwareDistributedVirtualSwitchTrunkVlanSpec'
        range = client_factory.create('ns0:NumericRange')
        range.start = start
        range.end = end
        vlan_tag = range
        vl_spec = client_factory.create(spec_ns)
        vl_spec.vlanId = vlan_tag
        vl_spec.inherited = '0'
        return vl_spec

    def _get_port_group_spec(self, net_id, vlan_tag, trunk_mode=False,
                             pg_spec=None):
        """Gets the port groups spec for net_id and vlan_tag."""
        client_factory = self._session.vim.client.factory
        if not pg_spec:
            pg_spec = client_factory.create('ns0:DVPortgroupConfigSpec')
        pg_spec.name = net_id
        pg_spec.type = 'ephemeral'
        config = client_factory.create('ns0:VMwareDVSPortSetting')
        if trunk_mode:
            config.vlan = self._get_trunk_vlan_spec()
        elif vlan_tag:
            config.vlan = self._get_vlan_spec(vlan_tag)

        pg_spec.defaultPortConfig = config
        return pg_spec

    def add_port_group(self, dvs_moref, net_id, vlan_tag=None,
                       trunk_mode=False):
        """Add a new port group to the configured DVS."""
        pg_spec = self._get_port_group_spec(net_id, vlan_tag,
                                            trunk_mode=trunk_mode)
        task = self._session.invoke_api(self._session.vim,
                                        'CreateDVPortgroup_Task',
                                        dvs_moref,
                                        spec=pg_spec)
        try:
            # NOTE(garyk): cache the returned moref
            self._session.wait_for_task(task)
        except Exception:
            # NOTE(garyk): handle more specific exceptions
            with excutils.save_and_reraise_exception():
                LOG.exception('Failed to create port group for '
                              '%(net_id)s with tag %(tag)s.',
                              {'net_id': net_id, 'tag': vlan_tag})
        LOG.info("%(net_id)s with tag %(vlan_tag)s created on %(dvs)s.",
                 {'net_id': net_id,
                  'vlan_tag': vlan_tag,
                  'dvs': dvs_moref.value})

    def _net_id_to_moref(self, dvs_moref, net_id):
        """Gets the moref for the specific neutron network."""
        # NOTE(garyk): return this from a cache if not found then invoke
        # code below.
        port_groups = self._session.invoke_api(vim_util,
                                               'get_object_properties',
                                               self._session.vim,
                                               dvs_moref,
                                               ['portgroup'])
        if len(port_groups) and hasattr(port_groups[0], 'propSet'):
            for prop in port_groups[0].propSet:
                for val in prop.val[0]:
                    props = self._session.invoke_api(vim_util,
                                                     'get_object_properties',
                                                     self._session.vim,
                                                     val, ['name'])
                    if len(props) and hasattr(props[0], 'propSet'):
                        for prop in props[0].propSet:
                            # match name or mor id
                            if net_id == prop.val or net_id == val.value:
                                # NOTE(garyk): update cache
                                return val
        raise exceptions.NetworkNotFound(net_id=net_id)

    def _is_vlan_network_by_moref(self, moref):
        """
        This can either be a VXLAN or a VLAN network. The type is determined
        by the prefix of the moref.
        """
        return moref.startswith(PORTGROUP_PREFIX)

    def _copy_port_group_spec(self, orig_spec):
        client_factory = self._session.vim.client.factory
        pg_spec = client_factory.create('ns0:DVPortgroupConfigSpec')
        pg_spec.autoExpand = orig_spec['autoExpand']
        pg_spec.configVersion = orig_spec['configVersion']
        pg_spec.defaultPortConfig = orig_spec['defaultPortConfig']
        pg_spec.name = orig_spec['name']
        pg_spec.numPorts = orig_spec['numPorts']
        pg_spec.policy = orig_spec['policy']
        pg_spec.type = orig_spec['type']
        return pg_spec

    def update_port_group_spec_qos(self, pg_spec, qos_data):
        port_conf = pg_spec.defaultPortConfig
        # Update the in bandwidth shaping policy
        # Note: openstack refers to the directions from the VM point of view,
        # while the NSX refers to the vswitch point of view.
        # so open stack egress is actually inShaping here.
        inPol = port_conf.inShapingPolicy
        if qos_data.egress.bandwidthEnabled:
            inPol.inherited = False
            inPol.enabled.inherited = False
            inPol.enabled.value = True
            inPol.averageBandwidth.inherited = False
            inPol.averageBandwidth.value = qos_data.egress.averageBandwidth
            inPol.peakBandwidth.inherited = False
            inPol.peakBandwidth.value = qos_data.egress.peakBandwidth
            inPol.burstSize.inherited = False
            inPol.burstSize.value = qos_data.egress.burstSize
        else:
            inPol.inherited = True

        outPol = port_conf.outShapingPolicy
        if qos_data.ingress.bandwidthEnabled:
            outPol.inherited = False
            outPol.enabled.inherited = False
            outPol.enabled.value = True
            outPol.averageBandwidth.inherited = False
            outPol.averageBandwidth.value = qos_data.ingress.averageBandwidth
            outPol.peakBandwidth.inherited = False
            outPol.peakBandwidth.value = qos_data.ingress.peakBandwidth
            outPol.burstSize.inherited = False
            outPol.burstSize.value = qos_data.ingress.burstSize
        else:
            outPol.inherited = True

        # Update the DSCP marking
        if (port_conf.filterPolicy.inherited or
            len(port_conf.filterPolicy.filterConfig) == 0 or
            len(port_conf.filterPolicy.filterConfig[
                0].trafficRuleset.rules) == 0):

            if qos_data.dscpMarkEnabled:
                # create the entire structure
                client_factory = self._session.vim.client.factory
                filter_rule = client_factory.create('ns0:DvsTrafficRule')
                filter_rule.description = DSCP_RULE_DESCRIPTION
                filter_rule.action = client_factory.create(
                    'ns0:DvsUpdateTagNetworkRuleAction')
                filter_rule.action.dscpTag = qos_data.dscpMarkValue
                # mark only incoming packets (openstack egress = nsx ingress)
                filter_rule.direction = QOS_IN_DIRECTION
                # Add IP any->any qualifier
                qualifier = client_factory.create(
                    'ns0:DvsIpNetworkRuleQualifier')
                qualifier.protocol = 0
                qualifier.sourceAddress = None
                qualifier.destinationAddress = None
                filter_rule.qualifier = [qualifier]

                traffic_filter_config = client_factory.create(
                    'ns0:DvsTrafficFilterConfig')
                traffic_filter_config.trafficRuleset.rules = [filter_rule]
                traffic_filter_config.trafficRuleset.enabled = True
                traffic_filter_config.agentName = QOS_AGENT_NAME
                traffic_filter_config.inherited = False

                port_conf.filterPolicy = client_factory.create(
                    'ns0:DvsFilterPolicy')
                port_conf.filterPolicy.filterConfig = [
                    traffic_filter_config]
                port_conf.filterPolicy.inherited = False
        else:
            # The structure was already initialized
            filter_policy = port_conf.filterPolicy
            if qos_data.dscpMarkEnabled:
                # just update the DSCP value
                traffic_filter_config = filter_policy.filterConfig[0]
                filter_rule = traffic_filter_config.trafficRuleset.rules[0]
                filter_rule.action.dscpTag = qos_data.dscpMarkValue
            else:
                # delete the filter policy data
                filter_policy.filterConfig = []

    def _reconfigure_port_group(self, pg_moref, spec_update_calback,
                                spec_update_data):
        # Get the current configuration of the port group
        pg_spec = self._session.invoke_api(vim_util,
                                           'get_object_properties',
                                           self._session.vim,
                                           pg_moref, ['config'])
        if len(pg_spec) == 0 or len(pg_spec[0].propSet[0]) == 0:
            LOG.error('Failed to get object properties of %s', pg_moref)
            raise nsx_exc.DvsNotFound(dvs=pg_moref)

        # Convert the extracted config to DVPortgroupConfigSpec
        new_spec = self._copy_port_group_spec(pg_spec[0].propSet[0].val)

        # Update the configuration using the callback & data
        spec_update_calback(new_spec, spec_update_data)

        # Update the port group configuration
        task = self._session.invoke_api(self._session.vim,
                                        'ReconfigureDVPortgroup_Task',
                                        pg_moref, spec=new_spec)
        try:
            self._session.wait_for_task(task)
        except Exception:
            LOG.error('Failed to reconfigure DVPortGroup %s', pg_moref)
            raise nsx_exc.DvsNotFound(dvs=pg_moref)

    # Update the dvs port groups config for a vxlan/vlan network
    # update the spec using a callback and user data
    def update_port_groups_config(self, dvs_id, net_id, net_moref,
                                  spec_update_calback, spec_update_data):
        is_vlan = self._is_vlan_network_by_moref(net_moref)
        if is_vlan:
            return self._update_net_port_groups_config(net_moref,
                                                       spec_update_calback,
                                                       spec_update_data)
        else:
            dvs_moref = self._get_dvs_moref_by_id(dvs_id)
            return self._update_vxlan_port_groups_config(dvs_moref,
                                                         net_id,
                                                         net_moref,
                                                         spec_update_calback,
                                                         spec_update_data)

    # Update the dvs port groups config for a vxlan network
    # Searching the port groups for a partial match to the network id & moref
    # update the spec using a callback and user data
    def _update_vxlan_port_groups_config(self,
                                         dvs_moref,
                                         net_id,
                                         net_moref,
                                         spec_update_calback,
                                         spec_update_data):
        port_groups = self._session.invoke_api(vim_util,
                                               'get_object_properties',
                                               self._session.vim,
                                               dvs_moref,
                                               ['portgroup'])
        found = False
        if len(port_groups) and hasattr(port_groups[0], 'propSet'):
            for prop in port_groups[0].propSet:
                for pg_moref in prop.val[0]:
                    props = self._session.invoke_api(vim_util,
                                                     'get_object_properties',
                                                     self._session.vim,
                                                     pg_moref, ['name'])
                    if len(props) and hasattr(props[0], 'propSet'):
                        for prop in props[0].propSet:
                            if net_id in prop.val and net_moref in prop.val:
                                found = True
                                self._reconfigure_port_group(
                                    pg_moref,
                                    spec_update_calback,
                                    spec_update_data)

        if not found:
            raise exceptions.NetworkNotFound(net_id=net_id)

    # Update the dvs port groups config for a vlan network
    # Finding the port group using the exact moref of the network
    # update the spec using a callback and user data
    def _update_net_port_groups_config(self,
                                       net_moref,
                                       spec_update_calback,
                                       spec_update_data):
        pg_moref = vim_util.get_moref(net_moref,
                                      "DistributedVirtualPortgroup")
        self._reconfigure_port_group(pg_moref,
                                     spec_update_calback,
                                     spec_update_data)

    def delete_port_group(self, dvs_moref, net_id):
        """Delete a specific port group."""
        moref = self._net_id_to_moref(dvs_moref, net_id)
        task = self._session.invoke_api(self._session.vim,
                                        'Destroy_Task',
                                        moref)
        try:
            self._session.wait_for_task(task)
        except Exception:
            # NOTE(garyk): handle more specific exceptions
            with excutils.save_and_reraise_exception():
                LOG.exception('Failed to delete port group for %s.',
                              net_id)
        LOG.info("%(net_id)s delete from %(dvs)s.",
                 {'net_id': net_id,
                  'dvs': dvs_moref.value})

    def get_port_group_info(self, dvs_moref, net_id):
        """Get portgroup information."""
        pg_moref = self._net_id_to_moref(dvs_moref, net_id)
        # Expand the properties to collect on need basis.
        properties = ['name']
        pg_info = self._session.invoke_api(vim_util,
                                           'get_object_properties_dict',
                                           self._session.vim,
                                           pg_moref, properties)
        return pg_info

    def _get_dvs_moref_from_teaming_data(self, teaming_data):
        """Get the moref dvs that belongs to the teaming data"""
        if 'switchObj' in teaming_data:
            if 'objectId' in teaming_data['switchObj']:
                dvs_id = teaming_data['switchObj']['objectId']
                return vim_util.get_moref(
                    dvs_id, 'VmwareDistributedVirtualSwitch')

    def update_port_group_spec_teaming(self, pg_spec, teaming_data):
        mapping = {'FAILOVER_ORDER': 'failover_explicit',
                   'ETHER_CHANNEL': 'loadbalance_ip',
                   'LACP_ACTIVE': 'loadbalance_ip',
                   'LACP_PASSIVE': 'loadbalance_ip',
                   'LACP_V2': 'loadbalance_ip',
                   'LOADBALANCE_SRCID': 'loadbalance_srcid',
                   'LOADBALANCE_SRCMAC': 'loadbalance_srcmac',
                   'LOADBALANCE_LOADBASED': 'loadbalance_loadbased'}
        dvs_moref = self._get_dvs_moref_from_teaming_data(teaming_data)
        port_conf = pg_spec.defaultPortConfig
        policy = port_conf.uplinkTeamingPolicy
        policy.inherited = False
        policy.policy.inherited = False
        policy.policy.value = mapping[teaming_data['teamingPolicy']]
        policy.uplinkPortOrder.inherited = False
        ports = teaming_data['failoverUplinkPortNames']
        policy.uplinkPortOrder.activeUplinkPort = ports
        # The standby port will be those not configure as active ones
        uplinks = self._session.invoke_api(vim_util,
                                           "get_object_property",
                                           self._session.vim,
                                           dvs_moref,
                                           "config.uplinkPortPolicy")
        # VC does not support LAG and normal uplinks. So need to check
        # if we need to configure standby links
        if set(ports) & set(uplinks.uplinkPortName):
            standby = list(set(uplinks.uplinkPortName) - set(ports))
            policy.uplinkPortOrder.standbyUplinkPort = standby

    def update_port_group_spec_name(self, pg_spec, name):
        pg_spec.name = name

    def update_port_group_spec_trunk(self, pg_spec, trunk_data):
        port_conf = pg_spec.defaultPortConfig
        port_conf.vlan = self._get_trunk_vlan_spec()

    def update_port_group_security_policy(self, pg_spec, status):
        policy = pg_spec.policy
        policy.securityPolicyOverrideAllowed = status

    def _update_port_security_policy(self, dvs_moref, port, status):
        client_factory = self._session.vim.client.factory
        ps = client_factory.create('ns0:DVPortConfigSpec')
        ps.key = port.portKey
        ps.operation = 'edit'
        policy = client_factory.create('ns0:DVSSecurityPolicy')
        bp = client_factory.create('ns0:BoolPolicy')
        bp.inherited = False
        bp.value = status
        policy.allowPromiscuous = bp
        policy.forgedTransmits = bp
        policy.inherited = False
        setting = client_factory.create('ns0:VMwareDVSPortSetting')
        setting.securityPolicy = policy
        ps.setting = setting
        task = self._session.invoke_api(self._session.vim,
                                        'ReconfigureDVPort_Task',
                                        dvs_moref,
                                        port=ps)
        try:
            self._session.wait_for_task(task)
            LOG.info("Updated port security status")
        except Exception as e:
            LOG.error("Failed to update port %s. Reason: %s",
                      port.key, e)


class VMManager(VCManagerBase):
    """Management class for VMs related VC tasks."""

    def get_vm_moref_obj(self, instance_uuid):
        """Get reference to the VM.
        The method will make use of FindAllByUuid to get the VM reference.
        This method finds all VM's on the backend that match the
        instance_uuid, more specifically all VM's on the backend that have
        'config_spec.instanceUuid' set to 'instance_uuid'.
        """
        vm_refs = self._session.invoke_api(
            self._session.vim,
            API_FIND_ALL_BY_UUID,
            self._session.vim.service_content.searchIndex,
            uuid=instance_uuid,
            vmSearch=True,
            instanceUuid=True)
        if vm_refs:
            return vm_refs[0]

    def get_vm_moref(self, instance_uuid):
        """Get reference to the VM.
        """
        vm_ref = self.get_vm_moref_obj(instance_uuid)
        if vm_ref:
            return vm_ref.value

    def get_vm_spec(self, vm_moref):
        vm_specs = self._session.invoke_api(vim_util,
                                            'get_object_properties',
                                            self._session.vim,
                                            vm_moref, ['network'])
        if vm_specs:
            return vm_specs[0]

    def _build_vm_spec_attach(self, neutron_port_id, port_mac,
                              nsx_net_id, device_type):
        # Code inspired by nova: _create_vif_spec
        client_factory = self._session.vim.client.factory
        vm_spec = client_factory.create('ns0:VirtualMachineConfigSpec')
        device_change = client_factory.create('ns0:VirtualDeviceConfigSpec')
        device_change.operation = "add"

        net_device = client_factory.create('ns0:' + device_type)
        net_device.key = -47
        net_device.addressType = "manual"
        # configure the neutron port id and mac
        net_device.externalId = neutron_port_id
        net_device.macAddress = port_mac
        net_device.wakeOnLanEnabled = True

        backing = client_factory.create(
            'ns0:VirtualEthernetCardOpaqueNetworkBackingInfo')
        # configure the NSX network Id
        backing.opaqueNetworkId = nsx_net_id
        backing.opaqueNetworkType = "nsx.LogicalSwitch"
        net_device.backing = backing

        connectable_spec = client_factory.create(
            'ns0:VirtualDeviceConnectInfo')
        connectable_spec.startConnected = True
        connectable_spec.allowGuestControl = True
        connectable_spec.connected = True

        net_device.connectable = connectable_spec

        device_change.device = net_device
        vm_spec.deviceChange = [device_change]

        return vm_spec

    def attach_vm_interface(self, vm_moref, neutron_port_id,
                            port_mac, nsx_net_id, device_type):
        new_spec = self._build_vm_spec_attach(
            neutron_port_id, port_mac, nsx_net_id, device_type)
        task = self._session.invoke_api(self._session.vim,
                                        'ReconfigVM_Task',
                                        vm_moref,
                                        spec=new_spec)
        try:
            self._session.wait_for_task(task)
            LOG.info("Updated VM moref %(moref)s spec - "
                     "attached an interface",
                     {'moref': vm_moref.value})
        except Exception as e:
            LOG.error("Failed to reconfigure VM %(moref)s spec: %(e)s",
                      {'moref': vm_moref.value, 'e': e})

    def _build_vm_spec_detach(self, device):
        """Builds the vif detach config spec."""
        # Code inspired by nova: get_network_detach_config_spec
        client_factory = self._session.vim.client.factory
        config_spec = client_factory.create('ns0:VirtualMachineConfigSpec')
        virtual_device_config = client_factory.create(
                                'ns0:VirtualDeviceConfigSpec')
        virtual_device_config.operation = "remove"
        virtual_device_config.device = device
        config_spec.deviceChange = [virtual_device_config]
        return config_spec

    def detach_vm_interface(self, vm_moref, device):
        new_spec = self._build_vm_spec_detach(device)
        task = self._session.invoke_api(self._session.vim,
                                        'ReconfigVM_Task',
                                        vm_moref,
                                        spec=new_spec)
        try:
            self._session.wait_for_task(task)
            LOG.info("Updated VM %(moref)s spec - detached an interface",
                     {'moref': vm_moref.value})
        except Exception as e:
            LOG.error("Failed to reconfigure vm moref %(moref)s: %(e)s",
                      {'moref': vm_moref.value, 'e': e})

    def get_vm_interfaces_info(self, vm_moref):
        hardware_devices = self._session.invoke_api(vim_util,
                                                    "get_object_property",
                                                    self._session.vim,
                                                    vm_moref,
                                                    "config.hardware.device")
        return hardware_devices

    def _get_device_port(self, device_id, mac_address):
        vm_moref = self.get_vm_moref_obj(device_id)
        hardware_devices = self.get_vm_interfaces_info(vm_moref)
        if not hardware_devices:
            return
        if hardware_devices.__class__.__name__ == "ArrayOfVirtualDevice":
            hardware_devices = hardware_devices.VirtualDevice
        for device in hardware_devices:
            if hasattr(device, 'macAddress'):
                if device.macAddress == mac_address:
                    return device.backing.port

    def update_port_security_policy(self, dvs_id, net_id, net_moref,
                                    device_id, mac_address, status):
        dvs_moref = self._get_dvs_moref_by_id(dvs_id)
        port = self._get_device_port(device_id, mac_address)
        if port:
            self._update_port_security_policy(dvs_moref, port, status)


class ClusterManager(VCManagerBase):
    """Management class for Cluster related VC tasks."""

    def _reconfigure_cluster(self, session, cluster, config_spec):
        """Reconfigure a cluster in vcenter"""
        try:
            reconfig_task = session.invoke_api(
                session.vim, "ReconfigureComputeResource_Task",
                cluster, spec=config_spec, modify=True)
            session.wait_for_task(reconfig_task)
        except Exception as excep:
            LOG.exception('Failed to reconfigure cluster %s', excep)

    def _create_vm_group_spec(self, client_factory, name, vm_refs,
                              group=None):
        if group is None:
            group = client_factory.create('ns0:ClusterVmGroup')
            group.name = name
            operation = 'add'
        else:
            operation = 'edit'

        # On vCenter UI, it is not possible to create VM group without
        # VMs attached to it. But, using APIs, it is possible to create
        # VM group without VMs attached. Therefore, check for existence
        # of vm attribute in the group to avoid exceptions
        if hasattr(group, 'vm'):
            group.vm += vm_refs
        else:
            group.vm = vm_refs

        group_spec = client_factory.create('ns0:ClusterGroupSpec')
        group_spec.operation = operation
        group_spec.info = group
        return [group_spec]

    def _create_cluster_rules_spec(self, client_factory, name, vm_group_name,
                                   host_group_name):
        rules_spec = client_factory.create('ns0:ClusterRuleSpec')
        rules_spec.operation = 'add'
        policy_class = 'ns0:ClusterVmHostRuleInfo'
        rules_info = client_factory.create(policy_class)
        rules_info.enabled = True
        rules_info.mandatory = False
        rules_info.name = name
        rules_info.vmGroupName = vm_group_name
        rules_info.affineHostGroupName = host_group_name
        rules_spec.info = rules_info
        return rules_spec

    def get_configured_vms(self, resource_id, n_host_groups=2):
        session = self._session
        resource = vim_util.get_moref(resource_id, 'ResourcePool')
        # TODO(garyk): cache the cluster details
        cluster = session.invoke_api(
            vim_util, "get_object_property", self._session.vim, resource,
            "owner")
        cluster_config = session.invoke_api(
            vim_util, "get_object_property", self._session.vim, cluster,
            "configurationEx")
        configured_vms = []
        for index in range(n_host_groups):
            vm_group = None
            entry_id = index + 1
            groups = []
            if hasattr(cluster_config, 'group'):
                groups = cluster_config.group
            for group in groups:
                if 'neutron-group-%s' % entry_id == group.name:
                    vm_group = group
                    break
            if vm_group and hasattr(vm_group, 'vm'):
                for vm in vm_group.vm:
                    configured_vms.append(vm.value)
        return configured_vms

    def update_cluster_edge_failover(self, resource_id, vm_moids,
                                     host_group_names):
        """Updates cluster for vm placement using DRS"""
        session = self._session
        resource = vim_util.get_moref(resource_id, 'ResourcePool')
        # TODO(garyk): cache the cluster details
        cluster = session.invoke_api(
            vim_util, "get_object_property", self._session.vim, resource,
            "owner")
        cluster_config = session.invoke_api(
            vim_util, "get_object_property", self._session.vim, cluster,
            "configurationEx")
        vms = [vim_util.get_moref(vm_moid, 'VirtualMachine')
               if vm_moid else None
               for vm_moid in vm_moids]
        client_factory = session.vim.client.factory
        config_spec = client_factory.create('ns0:ClusterConfigSpecEx')
        num_host_groups = len(host_group_names)

        rules = []
        if hasattr(cluster_config, 'rule'):
            rules = cluster_config.rule

        for index, vm in enumerate(vms, start=1):
            if not vm:
                continue
            vmGroup = None
            groups = []
            if hasattr(cluster_config, 'group'):
                groups = cluster_config.group
            for group in groups:
                if 'neutron-group-%s' % index == group.name:
                    vmGroup = group
                    break
            # Create/update the VM group
            groupSpec = self._create_vm_group_spec(
                            client_factory,
                            'neutron-group-%s' % index,
                            [vm], vmGroup)
            config_spec.groupSpec.append(groupSpec)
            config_rule = None
            # Create the config rule if it does not exist
            for rule in rules:
                if 'neutron-rule-%s' % index == rule.name:
                    config_rule = rule
                    break
            if config_rule is None and index <= num_host_groups:
                ruleSpec = self._create_cluster_rules_spec(
                    client_factory, 'neutron-rule-%s' % index,
                    'neutron-group-%s' % index,
                    host_group_names[index - 1])
                config_spec.rulesSpec.append(ruleSpec)
        self._reconfigure_cluster(session, cluster, config_spec)

    def validate_host_groups(self, resource_id, host_group_names):
        session = self._session
        resource = vim_util.get_moref(resource_id, 'ResourcePool')
        cluster = session.invoke_api(
            vim_util, "get_object_property", self._session.vim, resource,
            "owner")
        client_factory = session.vim.client.factory
        config_spec = client_factory.create('ns0:ClusterConfigSpecEx')
        cluster_config = session.invoke_api(
            vim_util, "get_object_property", self._session.vim, cluster,
            "configurationEx")
        groups = []
        if hasattr(cluster_config, 'group'):
            groups = cluster_config.group
        for host_group_name in host_group_names:
            found = False
            for group in groups:
                if host_group_name == group.name:
                    found = True
                    break
            if not found:
                LOG.error("%s does not exist", host_group_name)
                raise exceptions.NotFound()

        update_cluster = False
        num_host_groups = len(host_group_names)
        rules = []
        if hasattr(cluster_config, 'rule'):
            rules = cluster_config.rule
        # Ensure that the VM groups are created
        for index in range(num_host_groups):
            entry_id = index + 1
            vmGroup = None
            for group in groups:
                if 'neutron-group-%s' % entry_id == group.name:
                    vmGroup = group
                    break
            if vmGroup is None:
                groupSpec = self._create_vm_group_spec(
                                client_factory,
                                'neutron-group-%s' % entry_id,
                                [], vmGroup)
                config_spec.groupSpec.append(groupSpec)
                update_cluster = True

            config_rule = None
            # Create the config rule if it does not exist
            for rule in rules:
                if 'neutron-rule-%s' % entry_id == rule.name:
                    config_rule = rule
                    break
            if config_rule is None and index < num_host_groups:
                ruleSpec = self._create_cluster_rules_spec(
                    client_factory, 'neutron-rule-%s' % entry_id,
                    'neutron-group-%s' % entry_id,
                    host_group_names[index - 1])
                config_spec.rulesSpec.append(ruleSpec)
                update_cluster = True
        if update_cluster:
            try:
                self._reconfigure_cluster(session, cluster, config_spec)
            except Exception as e:
                LOG.error('Unable to update cluster for host groups %s', e)

    def _delete_vm_group_spec(self, client_factory, name):
        group_spec = client_factory.create('ns0:ClusterGroupSpec')
        group = client_factory.create('ns0:ClusterVmGroup')
        group.name = name
        group_spec.operation = 'remove'
        group_spec.removeKey = name
        group_spec.info = group
        return [group_spec]

    def _delete_cluster_rules_spec(self, client_factory, rule):
        rules_spec = client_factory.create('ns0:ClusterRuleSpec')
        rules_spec.operation = 'remove'
        rules_spec.removeKey = int(rule.key)
        policy_class = 'ns0:ClusterVmHostRuleInfo'
        rules_info = client_factory.create(policy_class)
        rules_info.name = rule.name
        rules_info.vmGroupName = rule.vmGroupName
        rules_info.affineHostGroupName = rule.affineHostGroupName
        rules_spec.info = rules_info
        return rules_spec

    def cluster_host_group_cleanup(self, resource_id, n_host_groups=2):
        session = self._session
        resource = vim_util.get_moref(resource_id, 'ResourcePool')
        # TODO(garyk): cache the cluster details
        cluster = session.invoke_api(
            vim_util, "get_object_property", self._session.vim, resource,
            "owner")
        client_factory = session.vim.client.factory
        config_spec = client_factory.create('ns0:ClusterConfigSpecEx')
        cluster_config = session.invoke_api(
            vim_util, "get_object_property", self._session.vim, cluster,
            "configurationEx")
        groups = []
        if hasattr(cluster_config, 'group'):
            groups = cluster_config.group
        rules = []
        if hasattr(cluster_config, 'rule'):
            rules = cluster_config.rule

        groupSpec = []
        ruleSpec = []
        for index in range(n_host_groups):
            entry_id = index + 1
            for group in groups:
                if 'neutron-group-%s' % entry_id == group.name:
                    groupSpec.append(self._delete_vm_group_spec(
                        client_factory, group.name))
            # Delete the config rule if it exists
            for rule in rules:
                if 'neutron-rule-%s' % entry_id == rule.name:
                    ruleSpec.append(self._delete_cluster_rules_spec(
                        client_factory, rule))

        if groupSpec:
            config_spec.groupSpec = groupSpec
        if ruleSpec:
            config_spec.rulesSpec = ruleSpec
        if groupSpec or ruleSpec:
            self._reconfigure_cluster(session, cluster, config_spec)


class VCManager(DvsManager, VMManager, ClusterManager):
    """Management class for all vc related tasks."""
    pass



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import mock
from openstackclient.tests.unit.network.v2 import (
    test_security_group_network as test_security_group)
from openstackclient.tests.unit import utils as tests_utils

from vmware_nsx.osc.v2 import security_group


supported_extensions = ('security-group-logging',
                        'provider-security-group',
                        'security-group-policy')


class TestCreateSecurityGroup(
    test_security_group.TestCreateSecurityGroupNetwork):

    def setUp(self):
        super(TestCreateSecurityGroup, self).setUp()
        # Get the command object to test
        self.cmd = security_group.NsxCreateSecurityGroup(
            self.app, self.namespace)
        # mock the relevant extensions
        get_ext = mock.patch('vmware_nsx.osc.v2.utils.get_extensions').start()
        get_ext.return_value = supported_extensions

    def _test_create_with_flag_arg(
        self, arg_name, validate_name, validate_val):
        self.network.create_security_group = mock.Mock(
            return_value=self._security_group)
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            '--description', self._security_group.description,
            conv_name,
            self._security_group.name
        ]
        verifylist = [
            ('description', self._security_group.description),
            ('name', self._security_group.name),
            (arg_name, True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network.create_security_group.assert_called_once_with(**{
            'description': self._security_group.description,
            'name': self._security_group.name,
            validate_name: validate_val,
        })
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_with_logging(self):
        self._test_create_with_flag_arg('logging', 'logging', True)

    def test_create_with_no_logging(self):
        self._test_create_with_flag_arg('no_logging', 'logging', False)

    def test_create_with_provider(self):
        self._test_create_with_flag_arg('provider', 'provider', True)

    def _test_create_with_arg_val(self, arg_name, arg_val):
        self.network.create_security_group = mock.Mock(
            return_value=self._security_group)
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            '--description', self._security_group.description,
            conv_name, str(arg_val),
            self._security_group.name
        ]
        verifylist = [
            ('description', self._security_group.description),
            ('name', self._security_group.name),
            (arg_name, arg_val),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network.create_security_group.assert_called_once_with(**{
            'description': self._security_group.description,
            'name': self._security_group.name,
            arg_name: arg_val,
        })
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_with_policy(self):
        self._test_create_with_arg_val('policy', 'policy-1')


class TestSetSecurityGroup(
    test_security_group.TestSetSecurityGroupNetwork):

    def setUp(self):
        super(TestSetSecurityGroup, self).setUp()
        # Get the command object to test
        self.cmd = security_group.NsxSetSecurityGroup(
            self.app, self.namespace)
        # mock the relevant extensions
        get_ext = mock.patch('vmware_nsx.osc.v2.utils.get_extensions').start()
        get_ext.return_value = supported_extensions

    def _test_set_with_flag_arg(self, arg_name, validate_name,
                                validate_val, is_valid=True):
        self.network.create_security_group = mock.Mock(
            return_value=self._security_group)
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            conv_name,
            self._security_group.name
        ]
        verifylist = [
            (arg_name, True),
            ('group', self._security_group.name),
        ]

        if not is_valid:
            self.assertRaises(tests_utils.ParserException, self.check_parser,
                              self.cmd, arglist, verifylist)
            return

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network.update_security_group.assert_called_once_with(
            self._security_group,
            **{validate_name: validate_val})
        self.assertIsNone(result)

    def test_set_with_logging(self):
        self._test_set_with_flag_arg('logging', 'logging', True)

    def test_set_with_no_logging(self):
        self._test_set_with_flag_arg('no_logging', 'logging', False)

    def test_set_with_provider(self):
        # modifying the provider flag should fail
        self._test_set_with_flag_arg('provider', 'provider',
                                     True, is_valid=False)

    def _test_set_with_arg_val(self, arg_name, arg_val):
        self.network.create_security_group = mock.Mock(
            return_value=self._security_group)
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            conv_name, str(arg_val),
            self._security_group.name
        ]
        verifylist = [
            (arg_name, arg_val),
            ('group', self._security_group.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network.update_security_group.assert_called_once_with(
            self._security_group,
            **{arg_name: arg_val})
        self.assertIsNone(result)

    def test_set_with_policyr(self):
        self._test_set_with_arg_val('policy', 'policy-1')



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock

from neutron_lib import context
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.common import exceptions
from neutron.objects import base as base_object
from neutron.objects.qos import policy as policy_object
from neutron.objects.qos import rule as rule_object
from neutron.services.qos import qos_plugin
from neutron.tests.unit.services.qos import base

from vmware_nsx.db import db as nsx_db
from vmware_nsx.plugins.nsx_v3 import utils as v3_utils
from vmware_nsx.services.qos.nsx_v3 import driver as qos_driver
from vmware_nsx.services.qos.nsx_v3 import utils as qos_utils
from vmware_nsx.tests.unit.nsx_v3 import test_plugin

PLUGIN_NAME = 'vmware_nsx.plugins.nsx_v3.plugin.NsxV3Plugin'


class TestQosNsxV3Notification(base.BaseQosTestCase,
                               test_plugin.NsxV3PluginTestCaseMixin):

    def setUp(self):
        # Reset the drive to re-create it
        qos_driver.DRIVER = None
        super(TestQosNsxV3Notification, self).setUp()
        self.setup_coreplugin(PLUGIN_NAME)

        self.qos_plugin = qos_plugin.QoSPlugin()
        self.ctxt = context.Context('fake_user', 'fake_tenant')
        mock.patch.object(self.ctxt.session, 'refresh').start()
        mock.patch.object(self.ctxt.session, 'expunge').start()
        self.policy_data = {
            'policy': {'id': uuidutils.generate_uuid(),
                       'project_id': uuidutils.generate_uuid(),
                       'name': 'test-policy',
                       'description': 'Test policy description',
                       'shared': True}}
        self.rule_data = {
            'bandwidth_limit_rule': {'id': uuidutils.generate_uuid(),
                                     'max_kbps': 2000,
                                     'max_burst_kbps': 150}}
        self.ingress_rule_data = {
            'bandwidth_limit_rule': {'id': uuidutils.generate_uuid(),
                                     'max_kbps': 3000,
                                     'max_burst_kbps': 350,
                                     'direction': 'ingress'}}
        self.dscp_rule_data = {
            'dscp_marking_rule': {'id': uuidutils.generate_uuid(),
                                  'dscp_mark': 22}}

        self.policy = policy_object.QosPolicy(
            self.ctxt, **self.policy_data['policy'])

        # egress BW limit rule
        self.rule = rule_object.QosBandwidthLimitRule(
            self.ctxt, **self.rule_data['bandwidth_limit_rule'])
        # ingress bw limit rule
        self.ingress_rule = rule_object.QosBandwidthLimitRule(
            self.ctxt, **self.ingress_rule_data['bandwidth_limit_rule'])
        self.dscp_rule = rule_object.QosDscpMarkingRule(
            self.ctxt, **self.dscp_rule_data['dscp_marking_rule'])

        self.fake_profile_id = 'fake_profile'
        self.fake_profile = {'id': self.fake_profile_id}

        mock.patch('neutron.objects.db.api.create_object').start()
        mock.patch('neutron.objects.db.api.update_object').start()
        mock.patch('neutron.objects.db.api.delete_object').start()
        mock.patch.object(nsx_db, 'get_switch_profile_by_qos_policy',
                          return_value=self.fake_profile_id).start()

        self.peak_bw_multiplier = cfg.CONF.NSX.qos_peak_bw_multiplier

        self.nsxlib = v3_utils.get_nsxlib_wrapper()

    @mock.patch(
        'neutron.objects.rbac_db.RbacNeutronDbObjectMixin'
        '.create_rbac_policy')
    @mock.patch.object(nsx_db, 'add_qos_policy_profile_mapping')
    def test_policy_create_profile(self, fake_db_add, fake_rbac_create):
        # test the switch profile creation when a QoS policy is created
        with mock.patch(
            'vmware_nsxlib.v3.core_resources.NsxLibQosSwitchingProfile.create',
            return_value=self.fake_profile
        ) as create_profile:
            with mock.patch('neutron.objects.qos.policy.QosPolicy.get_object',
                return_value=self.policy):
                with mock.patch('neutron.objects.qos.policy.QosPolicy.create'):
                    policy = self.qos_plugin.create_policy(self.ctxt,
                                                           self.policy_data)
                    expected_tags = self.nsxlib.build_v3_tags_payload(
                        policy,
                        resource_type='os-neutron-qos-id',
                        project_name=self.ctxt.tenant_name)

                    create_profile.assert_called_once_with(
                        description=self.policy_data["policy"]["description"],
                        name=self.policy_data["policy"]["name"],
                        tags=expected_tags)
                    # verify that the policy->profile mapping entry was added
                    self.assertTrue(fake_db_add.called)

    @mock.patch(
        'neutron.objects.rbac_db.RbacNeutronDbObjectMixin'
        '.create_rbac_policy')
    def __test_policy_update_profile(self, *mocks):
        # test the switch profile update when a QoS policy is updated
        fields = base_object.get_updatable_fields(
            policy_object.QosPolicy, self.policy_data['policy'])
        with mock.patch(
            'vmware_nsxlib.v3.core_resources.NsxLibQosSwitchingProfile.update'
        ) as update_profile:
            with mock.patch('neutron.objects.qos.policy.QosPolicy.get_object',
                return_value=self.policy):
                with mock.patch('neutron.objects.qos.policy.QosPolicy.update'):
                    self.qos_plugin.update_policy(
                        self.ctxt, self.policy.id, {'policy': fields})
                    # verify that the profile was updated with the correct data
                    self.policy_data["policy"]["id"] = self.policy.id
                    expected_tags = self.nsxlib.build_v3_tags_payload(
                        self.policy_data["policy"],
                        resource_type='os-neutron-qos-id',
                        project_name=self.ctxt.tenant_name)

                    update_profile.assert_called_once_with(
                        self.fake_profile_id,
                        description=self.policy_data["policy"]["description"],
                        name=self.policy_data["policy"]["name"],
                        tags=expected_tags
                    )

    @mock.patch.object(policy_object.QosPolicy, '_reload_rules')
    def test_bw_rule_create_profile(self, *mocks):
        # test the switch profile update when a egress QoS BW rule is created
        _policy = policy_object.QosPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [self.rule])
        with mock.patch('neutron.objects.qos.policy.QosPolicy.get_object',
            return_value=_policy):
            with mock.patch(
                'vmware_nsxlib.v3.core_resources.NsxLibQosSwitchingProfile.'
                'set_profile_shaping'
            ) as update_profile:
                with mock.patch('neutron.objects.db.api.update_object',
                    return_value=self.rule_data):
                    self.qos_plugin.update_policy_bandwidth_limit_rule(
                        self.ctxt, self.rule.id, _policy.id, self.rule_data)

                    # validate the data on the profile
                    rule_dict = self.rule_data['bandwidth_limit_rule']
                    expected_bw = int(round(float(
                        rule_dict['max_kbps']) / 1024))
                    expected_burst = rule_dict['max_burst_kbps'] * 128
                    expected_peak = int(expected_bw * self.peak_bw_multiplier)
                    # egress neutron rule -> ingress nsx args
                    update_profile.assert_called_once_with(
                        self.fake_profile_id,
                        ingress_bw_enabled=True,
                        ingress_burst_size=expected_burst,
                        ingress_peak_bandwidth=expected_peak,
                        ingress_average_bandwidth=expected_bw,
                        egress_bw_enabled=False,
                        egress_burst_size=None,
                        egress_peak_bandwidth=None,
                        egress_average_bandwidth=None,
                        dscp=0,
                        qos_marking='trusted'
                    )

    @mock.patch.object(policy_object.QosPolicy, '_reload_rules')
    def test_ingress_bw_rule_create_profile(self, *mocks):
        # test the switch profile update when a ingress QoS BW rule is created
        _policy = policy_object.QosPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [self.ingress_rule])
        with mock.patch('neutron.objects.qos.policy.QosPolicy.get_object',
            return_value=_policy):
            with mock.patch(
                'vmware_nsxlib.v3.core_resources.NsxLibQosSwitchingProfile.'
                'set_profile_shaping'
            ) as update_profile:
                with mock.patch('neutron.objects.db.api.update_object',
                    return_value=self.ingress_rule_data):
                    self.qos_plugin.update_policy_bandwidth_limit_rule(
                        self.ctxt, self.ingress_rule.id, _policy.id,
                        self.ingress_rule_data)

                    # validate the data on the profile
                    rule_dict = self.ingress_rule_data['bandwidth_limit_rule']
                    expected_bw = int(round(float(
                        rule_dict['max_kbps']) / 1024))
                    expected_burst = rule_dict['max_burst_kbps'] * 128
                    expected_peak = int(expected_bw * self.peak_bw_multiplier)
                    # ingress neutron rule -> egress nsx args
                    update_profile.assert_called_once_with(
                        self.fake_profile_id,
                        egress_bw_enabled=True,
                        egress_burst_size=expected_burst,
                        egress_peak_bandwidth=expected_peak,
                        egress_average_bandwidth=expected_bw,
                        ingress_bw_enabled=False,
                        ingress_burst_size=None,
                        ingress_peak_bandwidth=None,
                        ingress_average_bandwidth=None,
                        dscp=0,
                        qos_marking='trusted'
                    )

    @mock.patch.object(policy_object.QosPolicy, '_reload_rules')
    def test_bw_rule_create_profile_minimal_val(self, *mocks):
        # test driver precommit with an invalid limit value
        bad_limit = qos_utils.MAX_KBPS_MIN_VALUE - 1
        rule_data = {
            'bandwidth_limit_rule': {'id': uuidutils.generate_uuid(),
                                     'max_kbps': bad_limit,
                                     'max_burst_kbps': 150}}

        rule = rule_object.QosBandwidthLimitRule(
            self.ctxt, **rule_data['bandwidth_limit_rule'])

        _policy = policy_object.QosPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [rule])
        with mock.patch('neutron.objects.qos.policy.QosPolicy.get_object',
                        return_value=_policy),\
            mock.patch('neutron.objects.db.api.update_object',
                       return_value=rule_data):
            self.assertRaises(
                exceptions.DriverCallError,
                self.qos_plugin.update_policy_bandwidth_limit_rule,
                self.ctxt, rule.id, _policy.id, rule_data)

    @mock.patch.object(policy_object.QosPolicy, '_reload_rules')
    def test_bw_rule_create_profile_maximal_val(self, *mocks):
        # test driver precommit with an invalid burst value
        bad_burst = qos_utils.MAX_BURST_MAX_VALUE + 1
        rule_data = {
            'bandwidth_limit_rule': {'id': uuidutils.generate_uuid(),
                                     'max_kbps': 1025,
                                     'max_burst_kbps': bad_burst}}

        rule = rule_object.QosBandwidthLimitRule(
            self.ctxt, **rule_data['bandwidth_limit_rule'])

        _policy = policy_object.QosPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [rule])
        with mock.patch('neutron.objects.qos.policy.QosPolicy.get_object',
                        return_value=_policy),\
            mock.patch('neutron.objects.db.api.update_object',
                       return_value=rule_data):
                self.assertRaises(
                    exceptions.DriverCallError,
                    self.qos_plugin.update_policy_bandwidth_limit_rule,
                    self.ctxt, rule.id, _policy.id, rule_data)

    @mock.patch.object(policy_object.QosPolicy, '_reload_rules')
    def test_dscp_rule_create_profile(self, *mocks):
        # test the switch profile update when a QoS DSCP rule is created
        _policy = policy_object.QosPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [self.dscp_rule])
        with mock.patch('neutron.objects.qos.policy.QosPolicy.get_object',
            return_value=_policy):
            with mock.patch(
                'vmware_nsxlib.v3.core_resources.NsxLibQosSwitchingProfile.'
                'set_profile_shaping'
            ) as update_profile:
                with mock.patch('neutron.objects.db.api.'
                    'update_object', return_value=self.dscp_rule_data):
                    self.qos_plugin.update_policy_dscp_marking_rule(
                        self.ctxt, self.dscp_rule.id,
                        _policy.id, self.dscp_rule_data)

                    # validate the data on the profile
                    rule_dict = self.dscp_rule_data['dscp_marking_rule']
                    dscp_mark = rule_dict['dscp_mark']
                    update_profile.assert_called_once_with(
                        self.fake_profile_id,
                        ingress_bw_enabled=False,
                        ingress_burst_size=None,
                        ingress_peak_bandwidth=None,
                        ingress_average_bandwidth=None,
                        egress_bw_enabled=False,
                        egress_burst_size=None,
                        egress_peak_bandwidth=None,
                        egress_average_bandwidth=None,
                        dscp=dscp_mark,
                        qos_marking='untrusted'
                    )

    def test_rule_delete_profile(self):
        # test the switch profile update when a QoS rule is deleted
        _policy = policy_object.QosPolicy(
            self.ctxt, **self.policy_data['policy'])
        # The mock will return the policy without the rule,
        # as if it was deleted
        with mock.patch('neutron.objects.qos.policy.QosPolicy.get_object',
            return_value=_policy):
            with mock.patch(
                'vmware_nsxlib.v3.core_resources.NsxLibQosSwitchingProfile.'
                'set_profile_shaping'
            ) as update_profile:
                setattr(_policy, "rules", [self.rule])
                self.qos_plugin.delete_policy_bandwidth_limit_rule(
                    self.ctxt, self.rule.id, self.policy.id)
                # validate the data on the profile
                update_profile.assert_called_once_with(
                    self.fake_profile_id,
                    ingress_bw_enabled=False,
                    ingress_burst_size=None,
                    ingress_peak_bandwidth=None,
                    ingress_average_bandwidth=None,
                    egress_bw_enabled=False,
                    egress_burst_size=None,
                    egress_peak_bandwidth=None,
                    egress_average_bandwidth=None,
                    dscp=0,
                    qos_marking='trusted'
                )

    @mock.patch('neutron.objects.db.api.get_object', return_value=None)
    def test_policy_delete_profile(self, *mocks):
        # test the switch profile deletion when a QoS policy is deleted
        with mock.patch(
            'vmware_nsxlib.v3.core_resources.NsxLibQosSwitchingProfile.'
            'delete',
            return_value=self.fake_profile
        ) as delete_profile:
            self.qos_plugin.delete_policy(self.ctxt, self.policy.id)
            delete_profile.assert_called_once_with(self.fake_profile_id)



# Copyright 2015 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_l2gw.db.l2gateway import l2gateway_db
from networking_l2gw.db.l2gateway import l2gateway_models as models
from networking_l2gw.services.l2gateway.common import constants as l2gw_const
from networking_l2gw.services.l2gateway import exceptions as l2gw_exc
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from oslo_log import log as logging
from oslo_utils import uuidutils

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import nsxv_constants
from vmware_nsx.db import db as nsx_db
from vmware_nsx.db import nsxv_db
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.nsx_v import availability_zones as nsx_az
from vmware_nsx.plugins.nsx_v.vshield.common import exceptions

LOG = logging.getLogger(__name__)


class NsxvL2GatewayDriver(l2gateway_db.L2GatewayMixin):

    """Class to handle API calls for L2 gateway and NSXv backend."""

    def __init__(self, plugin):
        super(NsxvL2GatewayDriver, self).__init__()
        self._plugin = plugin
        self.__core_plugin = None

    @property
    def _core_plugin(self):
        if not self.__core_plugin:
            self.__core_plugin = directory.get_plugin()
            if self.__core_plugin.is_tvd_plugin():
                self.__core_plugin = self.__core_plugin.get_plugin_by_type(
                    projectpluginmap.NsxPlugins.NSX_V)
        return self.__core_plugin

    @property
    def _nsxv(self):
        return self._core_plugin.nsx_v

    @property
    def _edge_manager(self):
        return self._core_plugin.edge_manager

    def _validate_device_list(self, devices):
        # In NSX-v, one L2 gateway is mapped to one DLR.
        # So we expect only one device to be configured as part of
        # a L2 gateway resource.
        if len(devices) != 1:
            msg = _("Only a single device is supported for one L2 gateway")
            raise n_exc.InvalidInput(error_message=msg)

    def _get_l2gateway_interface(self, context, interface_name):
        """Get all l2gateway_interfaces_by interface_name."""
        session = context.session
        with session.begin():
            return session.query(models.L2GatewayInterface).filter_by(
                interface_name=interface_name).all()

    def _validate_interface_list(self, context, interfaces):
        # In NSXv, interface is mapped to a vDS VLAN port group.
        # Since HA is not supported, only one interface is expected
        if len(interfaces) != 1:
            msg = _("Only a single interface is supported for one L2 gateway")
            raise n_exc.InvalidInput(error_message=msg)
        if not self._nsxv.vcns.validate_network(interfaces[0]['name']):
            msg = _("Configured interface not found")
            raise n_exc.InvalidInput(error_message=msg)
        interface = self._get_l2gateway_interface(context,
                                                  interfaces[0]['name'])
        if interface:
            msg = _("%s is already used.") % interfaces[0]['name']
            raise n_exc.InvalidInput(error_message=msg)

    def create_l2_gateway_precommit(self, context, l2_gateway):
        pass

    def create_l2_gateway_postcommit(self, context, l2_gateway):
        pass

    def create_l2_gateway(self, context, l2_gateway):
        """Create a logical L2 gateway."""
        self._admin_check(context, 'CREATE')
        gw = l2_gateway[self.gateway_resource]
        devices = gw['devices']
        self._validate_device_list(devices)
        interfaces = devices[0]['interfaces']
        self._validate_interface_list(context, interfaces)
        # Create a dedicated DLR
        try:
            edge_id = self._create_l2_gateway_edge(context)
        except nsx_exc.NsxL2GWDeviceNotFound:
            LOG.exception("Failed to create backend device "
                          "for L2 gateway")
            raise

        devices[0]['device_name'] = edge_id
        l2_gateway[self.gateway_resource]['devices'] = devices
        return

    def update_l2_gateway_precommit(self, context, l2_gateway):
        pass

    def update_l2_gateway_postcommit(self, context, l2_gateway):
        pass

    def _create_l2_gateway_edge(self, context):
        # Create a dedicated DLR
        lrouter = {'name': nsxv_constants.L2_GATEWAY_EDGE,
                   'id': uuidutils.generate_uuid()}
        # Create the router on the default availability zone
        availability_zone = (nsx_az.NsxVAvailabilityZones().
            get_default_availability_zone())
        self._edge_manager.create_lrouter(context,
                                          lrouter, lswitch=None, dist=True,
                                          availability_zone=availability_zone)
        edge_binding = nsxv_db.get_nsxv_router_binding(context.session,
                                                       lrouter['id'])
        if not edge_binding:
            raise nsx_exc.NsxL2GWDeviceNotFound()
        # Enable edge HA on the DLR
        if availability_zone.edge_ha:
            edge_id = edge_binding['edge_id']
            self._edge_manager.nsxv_manager.update_edge_ha(edge_id)
        return edge_binding['edge_id']

    def _get_device(self, context, l2gw_id):
        devices = self._get_l2_gateway_devices(context, l2gw_id)
        return devices[0]

    def create_l2_gateway_connection_precommit(self, contex, gw_connection):
        pass

    def create_l2_gateway_connection_postcommit(self, context, gw_connection):
        network_id = gw_connection.get('network_id')
        virtual_wire = nsx_db.get_nsx_switch_ids(context.session, network_id)

        # In NSX-v, there will be only one device configured per L2 gateway.
        # The name of the device shall carry the backend DLR.
        l2gw_id = gw_connection.get(l2gw_const.L2GATEWAY_ID)
        device = self._get_device(context, l2gw_id)
        device_name = device.get('device_name')
        device_id = device.get('id')
        interface = self._get_l2_gw_interfaces(context, device_id)
        interface_name = interface[0].get("interface_name")
        bridge_name = "bridge-" + uuidutils.generate_uuid()
        bridge_dict = {"bridges":
                       {"bridge":
                        {"name": bridge_name,
                         "virtualWire": virtual_wire[0],
                         "dvportGroup": interface_name}}}
        try:
            self._nsxv.create_bridge(device_name, bridge_dict)
        except exceptions.VcnsApiException:
            LOG.exception("Failed to update NSX, "
                          "rolling back changes on neutron.")
            raise l2gw_exc.L2GatewayServiceDriverError(
                method='create_l2_gateway_connection_postcommit')
        return

    def create_l2_gateway_connection(self, context, l2_gateway_connection):
        """Create a L2 gateway connection."""
        gw_connection = l2_gateway_connection.get(l2gw_const.
                                                  CONNECTION_RESOURCE_NAME)
        l2gw_id = gw_connection.get(l2gw_const.L2GATEWAY_ID)
        gw_db = self._get_l2_gateway(context, l2gw_id)
        if gw_db.network_connections:
            raise nsx_exc.NsxL2GWInUse(gateway_id=l2gw_id)
        return

    def delete_l2_gateway_connection_precommit(self, context,
                                               l2_gateway_connection):
        pass

    def delete_l2_gateway_connection_postcommit(self, context,
                                                l2_gateway_connection):
        pass

    def delete_l2_gateway_connection(self, context, l2_gateway_connection):
        """Delete a L2 gateway connection."""
        self._admin_check(context, 'DELETE')
        gw_connection = self.get_l2_gateway_connection(context,
                                                       l2_gateway_connection)
        if not gw_connection:
            raise l2gw_exc.L2GatewayConnectionNotFound(
                l2_gateway_connection)
        l2gw_id = gw_connection.get(l2gw_const.L2GATEWAY_ID)
        device = self._get_device(context, l2gw_id)
        device_name = device.get('device_name')
        self._nsxv.delete_bridge(device_name)

    def delete_l2_gateway(self, context, l2_gateway):
        """Delete a L2 gateway."""
        self._admin_check(context, 'DELETE')
        device = self._get_device(context, l2_gateway)
        edge_id = device.get('device_name')
        rtr_binding = nsxv_db.get_nsxv_router_binding_by_edge(
                        context.session, edge_id)
        if rtr_binding:
            self._edge_manager.delete_lrouter(context,
                                              rtr_binding['router_id'])

    def delete_l2_gateway_precommit(self, context, l2_gateway):
        pass

    def delete_l2_gateway_postcommit(self, context, l2_gateway):
        pass



# Copyright 2013 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class TaskStatus(object):
    """Task running status.

    This is used by execution/status callback function to notify the
    task manager what's the status of current task, and also used for
    indication the final task execution result.
    """
    PENDING = 1
    COMPLETED = 2
    ERROR = 3
    ABORT = 4


class TaskState(object):
    """Current state of a task.

    This is to keep track of the current state of a task.
    NONE: the task is still in the queue
    START: the task is pull out from the queue and is about to be executed
    EXECUTED: the task has been executed
    STATUS: we're running periodic status check for this task
    RESULT: the task has finished and result is ready
    """
    NONE = -1
    START = 0
    EXECUTED = 1
    STATUS = 2
    RESULT = 3



# Copyright 2015 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import extensions
from neutron_lib import constants


ROUTER_TYPE = 'router_type'
VALID_TYPES = ['shared', 'exclusive']
EXTENDED_ATTRIBUTES_2_0 = {
    'routers': {
        ROUTER_TYPE: {'allow_post': True, 'allow_put': True,
                      'validate': {'type:values': VALID_TYPES},
                      'default': constants.ATTR_NOT_SPECIFIED,
                      'is_visible': True},
    }
}


class Routertype(extensions.ExtensionDescriptor):
    """Extension class supporting router type."""

    @classmethod
    def get_name(cls):
        return "Router Type"

    @classmethod
    def get_alias(cls):
        return "nsxv-router-type"

    @classmethod
    def get_description(cls):
        return "Enables configuration of NSXv router type."

    @classmethod
    def get_updated(cls):
        return "2015-1-12T10:00:00-00:00"

    def get_required_extensions(self):
        return ["router"]

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        return []

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright 2016 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Support shared pools with NSXv LBaaSv2 driver

Revision ID: d49ac91b560e
Revises: dbe29d208ac6
Create Date: 2016-07-21 05:03:35.369938

"""

from alembic import op
from sqlalchemy.engine import reflection

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = 'd49ac91b560e'
down_revision = 'dbe29d208ac6'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.NEWTON]


def upgrade():
    change_pk_constraint('nsxv_lbaas_pool_bindings', ['loadbalancer_id',
                                                      'pool_id'])
    change_pk_constraint('nsxv_lbaas_monitor_bindings', ['loadbalancer_id',
                                                         'pool_id',
                                                         'hm_id',
                                                         'edge_id'])


def change_pk_constraint(table_name, columns):
    inspector = reflection.Inspector.from_engine(op.get_bind())
    pk_constraint = inspector.get_pk_constraint(table_name)
    op.drop_constraint(pk_constraint.get('name'), table_name, type_='primary')
    op.drop_column(table_name, 'listener_id')
    op.create_primary_key(None, table_name, columns)



# Copyright 2013 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
import six

from vmware_nsx.common import nsx_utils

LOG = log.getLogger(__name__)
# Protocol number look up for supported protocols
protocol_num_look_up = {'tcp': 6, 'icmp': 1, 'udp': 17, 'ipv6-icmp': 58}


def _convert_to_nsx_rule(session, cluster, rule, with_id=False):
    """Converts a Neutron security group rule to the NSX format.

    This routine also replaces Neutron IDs with NSX UUIDs.
    """
    nsx_rule = {}
    params = ['remote_ip_prefix', 'protocol',
              'remote_group_id', 'port_range_min',
              'port_range_max', 'ethertype']
    if with_id:
        params.append('id')

    for param in params:
        value = rule.get(param)
        if param not in rule:
            nsx_rule[param] = value
        elif not value:
            pass
        elif param == 'remote_ip_prefix':
            nsx_rule['ip_prefix'] = rule['remote_ip_prefix']
        elif param == 'remote_group_id':
            nsx_rule['profile_uuid'] = nsx_utils.get_nsx_security_group_id(
                session, cluster, rule['remote_group_id'])

        elif param == 'protocol':
            try:
                nsx_rule['protocol'] = int(rule['protocol'])
            except (ValueError, TypeError):
                nsx_rule['protocol'] = (
                    protocol_num_look_up[rule['protocol']])
        else:
            nsx_rule[param] = value
    return nsx_rule


def _convert_to_nsx_rules(session, cluster, rules, with_id=False):
    """Converts a list of Neutron security group rules to the NSX format."""
    nsx_rules = {'logical_port_ingress_rules': [],
                 'logical_port_egress_rules': []}
    for direction in ['logical_port_ingress_rules',
                      'logical_port_egress_rules']:
        for rule in rules[direction]:
            nsx_rules[direction].append(
                _convert_to_nsx_rule(session, cluster, rule, with_id))
    return nsx_rules


def get_security_group_rules_nsx_format(session, cluster,
                                        security_group_rules, with_id=False):
    """Convert neutron security group rules into NSX format.

    This routine splits Neutron security group rules into two lists, one
    for ingress rules and the other for egress rules.
    """

    def fields(rule):
        _fields = ['remote_ip_prefix', 'remote_group_id', 'protocol',
                   'port_range_min', 'port_range_max', 'protocol', 'ethertype']
        if with_id:
            _fields.append('id')
        return dict((k, v) for k, v in six.iteritems(rule) if k in _fields)

    ingress_rules = []
    egress_rules = []
    for rule in security_group_rules:
        if rule.get('souce_group_id'):
            rule['remote_group_id'] = nsx_utils.get_nsx_security_group_id(
                session, cluster, rule['remote_group_id'])

        if rule['direction'] == 'ingress':
            ingress_rules.append(fields(rule))
        elif rule['direction'] == 'egress':
            egress_rules.append(fields(rule))
    rules = {'logical_port_ingress_rules': egress_rules,
             'logical_port_egress_rules': ingress_rules}
    return _convert_to_nsx_rules(session, cluster, rules, with_id)


def merge_security_group_rules_with_current(session, cluster,
                                            new_rules, current_rules):
    merged_rules = get_security_group_rules_nsx_format(
        session, cluster, current_rules)
    for new_rule in new_rules:
        rule = new_rule['security_group_rule']
        if rule['direction'] == 'ingress':
            merged_rules['logical_port_egress_rules'].append(
                _convert_to_nsx_rule(session, cluster, rule))
        elif rule['direction'] == 'egress':
            merged_rules['logical_port_ingress_rules'].append(
                _convert_to_nsx_rule(session, cluster, rule))
    return merged_rules


def remove_security_group_with_id_and_id_field(rules, rule_id):
    """Remove rule by rule_id.

    This function receives all of the current rule associated with a
    security group and then removes the rule that matches the rule_id. In
    addition it removes the id field in the dict with each rule since that
    should not be passed to nsx.
    """
    for rule_direction in rules.values():
        item_to_remove = None
        for port_rule in rule_direction:
            if port_rule['id'] == rule_id:
                item_to_remove = port_rule
            else:
                # remove key from dictionary for NSX
                del port_rule['id']
        if item_to_remove:
            rule_direction.remove(item_to_remove)
